from ._advmoves import WaltonAdvMove
from ._art import WaltonArt
from ._base import WaltonBase
from ._movements import WaltonMove
from ._polygon import WaltonPolygon


class Walton(WaltonAdvMove, WaltonArt, WaltonPolygon):
    """
    This is a diagnostic/demo class.
    It superposes molecules for merging to test a hypothesis.
    In the Frankenstein novel,
    (Captain Walton is told the story of the creation of Adam by Victor Frankenstein.
    Basically he is just a spectator but actually allows the story to be known:
    this class does nothing bar demonstrate what can be done)

    Given a list of mols (or smiles as named arguments via the ``from_smiles`` classmethod),
    It superposes them via the ``superpose_by_mcs`` or ``superpose_by_map`` methods.
    Calling the instance, will merge molecules and yield the merged molecule.
    The initial molecules are stored as a list in the attribute ``mols``.
    The merged molecule is stored in the attribute ``merged``.

    Walton has some methods to speed up operations

    .. code-block:: python

        demo = Walton.from_smiles(spiropentadiene='C1=CC21C=C2')
        demo.ring_on_plane(mol_idx=0, ring_idx=1)
        demo.print_coords(mol_idx=0)

    Walton is aimed at making life easier to set up an artificial merger condition and show it.

    .. code-block:: python

        demo = Walton.from_smiles(resorcinol='c1ccc(O)cc1O', eugenol='Oc1ccc(cc1OC)CC=C') # create instance
        demo.superpose_by_map({(0,1):{4:0, 3:1, 2:2}})  # superpose molecules by atom indices
        demo()  # merge (Fragmenstein's Monster)
        demo.show()  # show 2d, names and 3d

    The latter is equivalent to:

    .. code-block:: python

        display(demo)
        demo.to_nglview()

    The key feature of the class is that it allows easy transformations

    .. code-block:: python

    demo = Walton.from_smiles(furan='c1ccco1')
    display(demo)  # find atom numbers w/o counting
    demo.ring_on_plane(mol_idx=0)  # flat on xy
    demo.atom_on_axis(mol_idx=0, atom_idx=4, axis='x') # ox on axis
    # the atom_idx can be actually be passed a Geometry.Point3D:
    demo.atom_on_axis(mol_idx=0,
                     atom_idx=demo.get_centroid_of_atoms(1, 2, mol_idx=0),
                    axis='x')
    """



from typing import (Optional, Union, Tuple)

import numpy as np
from rdkit import Geometry

from ._movements import WaltonMove


class WaltonAdvMove(WaltonMove):

    def atom_on_plane(self, mol_idx: int, atom_idx: Union[int, Geometry.Point3D], plane: str):
        """
        Rotate the mol at ``mol_idx`` (``int``) in the attribite ``mols`` (``List[Chem.Mol]``)
        so that atom at ``atom_idx`` is on a plane (eg. 'xy', 'yz' or 'xz').

        It does so by getting the angle subtended by the position on the axis
        that is not on that plane and rotating it away.

        Say, a point on the plane of the axes x and y have a z position of zero.
        A point that is not can be rotated in the opposite direction of
        the angle between the z position (the opposite leg)
        and either the x or y position (the adjecent leg) around the third axis (y or x),
        thus removing the z position.
        The angle in radians is the inverse tangent of
        the ratio of the opposite over the adjencent.
        """
        coords: Geometry.Point3D = self.get_point(atom_idx, mol_idx)
        plane = plane.lower()
        assert plane in ['xy', 'xz', 'yx', 'yz', 'zx', 'zy']
        axis = plane[1]
        annulled = (set('xyz') - set(plane)).pop()
        opposite = getattr(coords, annulled)
        adjencent = getattr(coords, plane[0])
        if adjencent != 0:  # normal
            theta = -np.arctan(opposite / adjencent)
            self.rotate(mol_idx=mol_idx,
                        theta=theta,
                        axis=axis,
                        degrees=False)
        elif getattr(coords, plane[1]) != 0:  # one is already flat
            self.atom_on_plane(mol_idx, coords, plane[::-1])
        else:  # this is likely the origin
            pass

    def atom_on_axis(self, mol_idx: int, atom_idx: Union[int, Geometry.Point3D], axis: str):
        """

        Rotate the mol at ``mol_idx`` (``int``) in the attribite ``mols`` (``List[Chem.Mol]``)
        so that atom at ``atom_idx`` is on a plane (eg. 'xy', 'yz' or 'xz').
        This calls ``atom_on_plane`` as an axis in on two axis planes
        """
        axis_on_planes = {'x': ('xy', 'xz'),
                          'y': ('xy', 'yz'),
                          'z': ('xz', 'yz'),
                          }
        assert axis in axis_on_planes, f'Axis {axis} not in {axis_on_planes}'
        for plane in axis_on_planes[axis]:
            self.atom_on_plane(mol_idx, atom_idx, plane)

    def atom_to_origin(self, mol_idx: int, atom_idx: Union[int, Geometry.Point3D]):
        """
        Translate to origin.
        """
        coords: Geometry.Point3D = self.get_point(atom_idx, mol_idx)
        self.translate_by_point(mol_idx=mol_idx, point=coords, scale=-1)

    def translate_by_point(self, mol_idx: int, point: Geometry.Point3D, scale: float = +1):
        """
        Translate by the vector point (``rdkit.Geometry.Point3D``).
        Setting scale to -1 will subtract the point as seen
        """
        self.translate(mol_idx=mol_idx,
                       x=scale * point.x,
                       y=scale * point.y,
                       z=scale * point.z)

    def translate_parallel(self,
                           mol_idx: int,
                           distance: float,
                           base_atom_idx: Union[int, Geometry.Point3D],
                           pointer_atom_idx: Union[int, Geometry.Point3D],
                           ref_mol_idx: Optional[int] = None):
        if ref_mol_idx is None:
            ref_mol_idx = mol_idx
        from_coord: Geometry.Point3D = self.get_point(base_atom_idx, ref_mol_idx)
        to_coord: Geometry.Point3D = self.get_point(pointer_atom_idx, ref_mol_idx)
        # This is a unitary vector
        direction: Geometry.Point3D = from_coord.DirectionVector(to_coord)
        self.translate_by_point(mol_idx=mol_idx, point=direction, scale=distance)

    def ring_on_plane(self, mol_idx: int, ring_idx: int = 0, plane: str = 'yx', centroid_to_origin: bool = True):
        """
        Place the ring flat on a plane.
        The maths is shoddy and as a result there may be a bit of offset.

        centroid_to_origin=False may not fully keep the centroid in place.
        """
        # move to centroid for easy maths
        original_centroid = self.get_centroid_of_ring(ring_idx=ring_idx, mol_idx=mol_idx)
        self.atom_to_origin(mol_idx=mol_idx, atom_idx=original_centroid)  # first atom to origin
        for i in self.get_mol(mol_idx).GetRingInfo().AtomRings()[ring_idx]:
            self.atom_on_plane(mol_idx=mol_idx,
                               atom_idx=i,
                               plane=[plane, plane[::-1]][i % 2]  # I am not sure this is needed.
                               )
            centroid = self.get_centroid_of_ring(ring_idx=ring_idx, mol_idx=mol_idx)
            self.atom_to_origin(mol_idx=mol_idx, atom_idx=centroid)  # first atom to origin
        if not centroid_to_origin:
            self.translate_by_point(mol_idx=mol_idx, point=original_centroid, scale=1)

    def flatten_trio(self, mol_idx: int, atom_idcs: Tuple[int, int, int],
                     primary_axis: str = 'x',
                     secondary_axis: str = 'y',
                     first_to_origin: bool = True,
                     ):
        """
        Give three atom indices place the first and third on the primary axis and
        the second on the plane with a second axis.

        Say there's a propane and the axes are 1ary=x and 2ary=y then this method
        will place the triangle along the x axis, flat on the xy plane, with no z elevation.
        """
        original_origin = self.get_point(mol_idx=mol_idx, atom_idx=atom_idcs[0])
        self.atom_to_origin(mol_idx=mol_idx, atom_idx=atom_idcs[0])  # first atom to origin
        self.atom_on_plane(mol_idx=mol_idx, atom_idx=atom_idcs[2],
                           plane=primary_axis + secondary_axis)  # atom 2 flat (z=0)
        self.atom_on_plane(mol_idx=mol_idx, atom_idx=atom_idcs[1],
                           plane=secondary_axis + primary_axis)  # atom 1 flat (z=0)
        self.atom_on_axis(mol_idx=mol_idx, atom_idx=atom_idcs[2], axis=primary_axis)  # atom 2 on x axis
        if not first_to_origin:
            self.translate_by_point(mol_idx=mol_idx, point=original_origin, scale=1)

    # ----- quicker coord getters --------------------------------------------------

    def print_coords(self, mol_idx: int = 0, no_hydrogens: bool = True):
        for atom in self.get_mol(mol_idx).GetAtoms():  #: Chem.Atom  # noqa
            if atom.GetAtomicNum() == 1 and no_hydrogens:
                continue  # no hydrogens
            atom_idx: int = atom.GetIdx()
            coords = self.get_point(atom_idx, mol_idx)
            print(f'atom {atom_idx: >2}: x={coords.x:.1f} y={coords.y:.1f} z={coords.z:.1f}')  # legit print

    def get_centroid_of_atoms(self,
                              *idx_or_points: Union[int, Geometry.Point3D],
                              mol_idx: int = -1) -> Geometry.Point3D:
        """
        centroid of atoms is the euclidean mean of the atoms

        if a whole ring is wanted see ``get_centroid_of_ring``
        """
        points = [self.get_point(ip, mol_idx) for ip in idx_or_points]
        mean_coord: Callable[[int], float] = lambda ai: float(np.mean([p[ai] for p in points]))  # noqa
        return Geometry.Point3D(mean_coord(0), mean_coord(1), mean_coord(2))

    def get_centroid_of_ring(self, mol_idx: int, ring_idx: int = 0) -> Geometry.Point3D:
        """
        centroid of ring is the euclidean mean of the ring's atoms

        if only some atoms are wanted see ``get_centroid_of_atoms``
        """
        return self.get_centroid_of_atoms(*self.get_mol(mol_idx).GetRingInfo().AtomRings()[ring_idx], mol_idx=mol_idx)

    def get_ring_radius(self, mol_idx: int, ring_idx: int = 0) -> float:
        """
        Get the distance to the centroid to the first atom in the ring.
        """
        atoms = self.get_mol(mol_idx).GetRingInfo().AtomRings()[ring_idx]
        centroid = self.get_centroid_of_atoms(*atoms, mol_idx=mol_idx)
        return atoms[0].Distance(centroid)



"""
Do note that monster contains a few visualisation functions.

"""

from io import StringIO
from fragmenstein.branding import divergent_colors
from IPython.display import display, HTML
from rdkit import Chem
import os, re
from typing import (Optional)
from ._victor_common import _VictorCommon
from ..display import MolNGLWidget, patched_3Dmol_view, DISPLAYMODE  # real or mock
# # MolNGLWidget is a subclass of NGLWidget while patched_3Dmol_view is py3Dmol.view but with monkey-patches

class _VictorShow(_VictorCommon):
    # partial move out of utility module

    def to_nglview(self, print_legend: bool = False) -> MolNGLWidget:
        """`
        generates a NGLWidget (``IPython.display.display`` will show it)
        with the compounds and the merged if present.
        To override the colours:
        The colours will be those in the ``Chem.Mol``'s property ``_color`` if present.

        Returns -> nv.NGLWidget subclass
        """
        view = self.monster.to_nglview(show_positioned_mol=False)
        for molblock in (self.minimized_pdbblock, self.unminimized_pdbblock):
            if molblock is None:
                continue
            fh = StringIO(molblock)
            comp: ComponentViewer = view.add_component(fh, ext='pdb')  # noqa it's there.
            comp.add_representation('ball+stick', colorValue='white', multipleBond=True,
                                    sele=f'[{self.ligand_resn}]')
            # force it.
            comp.update_ball_and_stick(colorValue='white', multipleBond=True)
            view._js(f"""const comp = this.stage.compList[{len(self.hits)}]
                                 const target_sele = new NGL.Selection('[{self.ligand_resn}]');
                                 const radius = 5;
                                 const neigh_atoms = comp.structure.getAtomSetWithinSelection( target_sele, radius );
                                 const resi_atoms = comp.structure.getAtomSetWithinGroup( neigh_atoms );
                                 comp.addRepresentation( "line", {{sele: resi_atoms.toSeleString(), 
                                                                   colorValue: "gainsboro",
                                                                   multipleBond: true}} );
                                 comp.addRepresentation( "contact", {{sele: resi_atoms.toSeleString()}});
                            """)
            break
        else:
            pass
        # async madness: disabled for now.
        # view.center(f'[{self.ligand_resn}]')
        if print_legend:
            legend = self.monster.get_legend(show_positioned_mol=False)
            legend += ' Fragmenstein monster (white)'
            display(HTML(legend))
        return view

    def to_3Dmol(self, print_legend: bool = False) -> patched_3Dmol_view:
        view = self.monster.to_3Dmol(show_positioned_mol=False)
        view.add_template(self.apo_pdbblock)
        if print_legend:
            legend = self.monster.get_legend(show_positioned_mol=False)
            legend += ' Fragmenstein monster (white)'
            display(HTML(legend))
        return view


    # =================== save  ========================================================================================

    def make_pse(self,
                 filename: str = 'combo.pse',
                 extra_mols: Optional[Chem.Mol] = None):
        """
        Save a pse in the relevant folder. This is the Victor one.
        """
        assert '.pse' in filename, f'{filename} not .pse file'
        if extra_mols is None:
            extra_mols = []
        # ------------------------
        import pymol2
        with pymol2.PyMOL() as pymol:
            for hit in self.hits:
                hit_name = hit.GetProp('_Name')
                pymol.cmd.read_molstr(Chem.MolToMolBlock(hit), hit_name)
                if self.monster is None:
                    pymol.cmd.color('grey50', f'element C and {hit_name}')
                elif hit_name in self.monster.unmatched:
                    pymol.cmd.color('black', f'element C and {hit_name}')
                else:
                    pymol.cmd.color('white', f'element C and {hit_name}')
            if self.monster is not None and self.monster.positioned_mol is not None:
                pymol.cmd.read_molstr(Chem.MolToMolBlock(self.monster.positioned_mol), 'placed')
                pymol.cmd.color('magenta', f'element C and placed')
                pymol.cmd.zoom('byres (placed expand 4)')
                pymol.cmd.show('line', 'byres (placed around 4)')
            if self.minimized_mol is not None:
                pymol.cmd.read_molstr(Chem.MolToMolBlock(self.minimized_mol), 'minimised')
                pymol.cmd.color('green', f'element C and minimised')
            if self.minimized_pdbblock is not None:
                pymol.cmd.read_pdbstr(self.minimized_pdbblock, 'min_protein')
                pymol.cmd.color('gray50', f'element C and min_protein')
                pymol.cmd.hide('sticks', 'min_protein')
            if self.unminimized_pdbblock is not None:
                pymol.cmd.read_pdbstr(self.unminimized_pdbblock, 'unmin_protein')
                pymol.cmd.color('gray20', f'element C and unmin_protein')
                pymol.cmd.hide('sticks', 'unmin_protein')
                pymol.cmd.disable('unmin_protein')
            for mol in extra_mols:
                name = mol.GetProp('_Name')
                pymol.cmd.read_molstr(Chem.MolToMolBlock(mol, kekulize=False), name)
                pymol.cmd.color('magenta', f'{name} and name C*')
            pymol.cmd.save(os.path.join(self.work_path, self.long_name, filename))

    def make_steps_pse(self, filename: str = 'step.pse'):
        """
        Saves the steps in a pse file. For a more exhaustive file,
        use `make_pse` in Monster.
        """
        import pymol2
        assert '.pse' in filename, f'{filename} not .pse file'
        with pymol2.PyMOL() as pymol:
            for hit in self.hits:
                pymol.cmd.read_molstr(Chem.MolToMolBlock(hit, kekulize=False), hit.GetProp('_Name'))
            for label, mod in self.modifications:
                pymol.cmd.read_molstr(Chem.MolToMolBlock(mod, kekulize=False), re.sub('[^\w_]', '_', label))
            pymol.cmd.save(os.path.join(self.work_path, self.long_name, filename))

    # ===== MF keeps calling these from victor...

    def draw_nicely(self, *args, **kwargs):
        self.journal.warning('The `draw_nicely` method is in monster. Please use Monster.draw_nicely')
        self.monster.draw_nicely(*args, **kwargs)

    def show_comparison(self, *args, **kwargs):
        self.journal.warning('The `show_comparison` method is in monster. Please use Monster.show_comparison')
        self.monster.show_comparison(*args, **kwargs)





import operator
from typing import Optional, Dict, List, Tuple, Set, Sequence, Unpack, Union, Any, Iterable  # noqa cf. .legacy monkeypatch

from rdkit import Chem
from rdkit.Chem import AllChem

from ._refine import _MonsterRefine
from ..mcs_mapping import IndexMap, flip_mapping
from ..unmerge_mapper import Unmerge
from collections import Counter
from ...error import DistanceError, PoisonError

class _MonsterNone(_MonsterRefine):

    def no_blending(self, broad=False) -> None:
        """
        no merging is done. The hits are mapped individually. Not great for small fragments.
        """
        maps: Dict[str, List[Dict[int, int]]] = self._compute_maps(broad)
        # there is no primary hit in no_blending mode.
        unmerger = self._perform_unmerge(maps, n_poisonous=3 if broad else 0)
        self.unmatched = [m.GetProp('_Name') for m in unmerger.disregarded]
        self.positioned_mol, self.mol_options = self._place_unmerger(unmerger)


    def _perform_unmerge(self,
                         maps: Dict[str, List[Dict[int, int]]],
                         n_poisonous:int,
                         primary_name:Optional[str]=None) -> Unmerge:
        """
        The second third of the no_blending method.
        But also used by expansion mapping.

        :param maps:
        :return:
        """
        positive: Dict[str, List[Dict[int, int]]] = self._remove_negatives(maps)
        flipped_maps: Dict[str, List[Dict[int, int]]] =\
            {name: [flip_mapping(hm) for hm in hit_mappings] for name, hit_mappings in maps.items()}
        # todo flip round Unmerge. Unmerge wants name to dict of followup to hit... which is backwards.
        unmerger = Unmerge(followup=self.initial_mol,
                             mols=self.hits,
                             maps=flipped_maps,
                             no_discard=False)  # self.throw_on_discard?
        unmatched = [m.GetProp('_Name') for m in unmerger.disregarded]
        if n_poisonous and unmatched:
            self._remove_poisonous(flipped_maps, unmerger.poisonous_indices, n_poisonous, primary_name)
            retried_unmerger = Unmerge(followup=self.initial_mol,
                               mols=self.hits,
                               maps=flipped_maps,
                               no_discard=False)  # self.throw_on_discard?
            retried_unmatched = [m.GetProp('_Name') for m in retried_unmerger.disregarded]
            if len(retried_unmatched) < len(unmatched):
                self.journal.debug(f'Retried unmerge yield the matching of ' +
                                   f'{set(unmatched) - set(retried_unmatched)} ' +
                                   f'via the removal of {unmerger.poisonous_indices}')
                unmerger = retried_unmerger
                unmatched = retried_unmatched
        if not self.throw_on_discard:
            pass
        elif len(unmatched) == 0:
            pass
        elif len(unmerger.poisonous_indices):
            raise PoisonError(mol=unmatched, indices=unmerger.poisonous_indices)
        else:
            raise DistanceError(hits=unmatched)
        self.journal.debug(f'followup to scaffold {unmerger.combined_map}')
        return unmerger

    def _remove_poisonous(self,
                          flipped_maps,
                          poisonous_indices: List[int],
                          n_poisonous: int,
                          primary_name:Optional[str]=None) -> None:
        """
        By poisonous it is intended the few atoms that cause a mapping for placement to go wrong.

        :param flipped_maps: modified in place
        :param poisonous_indices: a list with repeated indices as unmerge.poisonous_indices: for commonness
        :param n_poisonous: max indices to remove
        :param primary_name: remove poisonous atoms from primary hit only
        :return:
        """
        for i, c in Counter(poisonous_indices).most_common(n_poisonous):  # followup index
            for name, followup2hits in flipped_maps.items():
                if primary_name is not None and name != primary_name:
                    continue
                for followup2hit in followup2hits:
                    if i not in followup2hit.keys():
                        continue
                    if name in self.custom_map and i in self.custom_map[name].values():
                        continue
                    self.journal.debug(f'Removing poisonous mapping: ' +
                                       f'hit {name} index {followup2hit[i]} to followup index {i}')
                    del followup2hit[i]

    def _place_unmerger(self, unmerger: Unmerge) -> Tuple[Chem.Mol, List[Chem.Mol]]:
        # ------------------ places the atoms with known mapping ------------------
        placed = self.place_from_map(target_mol=self.initial_mol,
                                     template_mol=unmerger.combined_bonded,
                                     atom_map=unmerger.combined_map,
                                     random_seed=self.random_seed)

        self.keep_copy(unmerger.combined, 'scaffold')
        self.keep_copy(unmerger.combined_bonded, 'chimera')

        alts = zip(unmerger.combined_bonded_alternatives, unmerger.combined_map_alternatives)
        placed_options = [self.place_from_map(target_mol=self.initial_mol,
                                             template_mol=mol,
                                             atom_map=mappa,
                                             random_seed=self.random_seed) for mol, mappa in alts]
        # ------------------ .posthoc_refine averages the overlapping atoms ------------------
        if unmerger.pick != -1:  # override present!
            self.journal.debug(f'override present: {unmerger.pick}')
            positioned_mol = self.posthoc_refine(placed)
            mol_options = [self.posthoc_refine(mol) for mol in placed_options]
        else:
            mols: List[Chem.Mol] = [self.posthoc_refine(mol) for mol in [placed] + placed_options]
            positioned_mol: Chem.Mol = self.get_best_scoring(mols)
            mols.remove(positioned_mol)
            mol_options: List[Chem.Mol] = mols
        return positioned_mol, mol_options

    def _compute_maps(self, broad: bool) -> Dict[str, List[Dict[int, int]]]:
        """
        Compute the mapping for each hit and returns a dict of name to list of mappings
        in the scheme of ``[{hit_atom_idx: template_atom_idx}, ...]``

        :param broad:
        :return:
        """
        return {hit.GetProp('_Name'): self._compute_hit_maps(hit, broad) for hit in self.hits}

    def _compute_hit_maps(self, template: Chem.Mol, broad: bool) -> List[Dict[int, int]]:
        """
        Calcualte the list of options of possible maps of hit
        This is called in the expansion tactic in ``_get_primary_maps``  (called to find the primary)
        indirectly via ``_compute_maps``
        or directly when a primary is known. The next step, does not call it but calls
        ``_expand_hit_atom_map_by_overlap``.

        :param template:
        :param broad:
        :return:
        """
        if broad:
            self.journal.debug('Merge ligands: False. Broad search: True')
            pair_atom_maps, _ = self.get_mcs_mappings(template, self.initial_mol)
            return pair_atom_maps
        else:
            self.journal.debug('Merge ligands: False. Broad search: False')
            pair_atom_maps_t: List[IndexMap] = self._get_atom_maps(followup=self.initial_mol, hit=template,
                                                                   **self.strict_matching_mode)
            pair_atom_maps: List[Dict[int, int]] = [dict(p) for p in pair_atom_maps_t]
            return pair_atom_maps

    def _remove_negatives(self, map: Any) -> Any:
        """
        This is a hack to remove negative values from the map.
        It like flip_mapping is a tangle of options. The latter has been typing.overload'ed

        :param map:
        :return:
        """
        if len(map) == 0:  # nothing to do
            return map
        # list of something
        if isinstance(map, list) and isinstance(map[0], tuple) and isinstance(map[0][0], tuple):
            return [(h, f) for h, f in map if f >= 0 and h >= 0]
        elif isinstance(map, list) and isinstance(map[0], list):
            return [self._remove_negatives(m) for m in map]
        elif isinstance(map, list) and isinstance(map[0], dict):
            return [self._remove_negatives(m) for m in map]
        elif isinstance(map, list):
            raise ValueError('Unsupported sub map type: {}'.format(type(map[0])))
        # dict of something
        if not isinstance(map, dict):
            raise ValueError('Unsupported map type: {}'.format(type(map)))
        first_val = list(map.values())[0]
        if isinstance(map, dict) and isinstance(first_val, int):
            return {h: f for h, f in map.items() if f >= 0 and h >= 0}
        elif isinstance(map, dict) and isinstance(first_val, Iterable):
            return {k: self._remove_negatives(v) for k, v in map.items()}
        else:
            raise ValueError('Unsupported sub map type: {}'.format(type(first_val)))



from rdkit.Chem import rdmolops

import json
from typing import Optional, Dict, List, Tuple, Set, Unpack, Union  # noqa cf. .legacy monkeypatch
from warnings import warn
from collections import defaultdict
import numpy as np
from rdkit import Chem
from rdkit.Chem import rdMolAlign
from rdkit.Geometry.rdGeometry import Point3D
from rdkit.Chem import AllChem
import operator

from ._make_chimera import _MonsterChimera


class _MonsterRefine(_MonsterChimera):

    def posthoc_refine(self,
                       scaffold: Chem.Mol,
                       indices: Optional[List[int]] = None) -> Chem.Mol:
        """
        Given a scaffold and a list of indices, refine the scaffold.

        :param scaffold:
        :param indices: if absent, use all atoms
        :return:
        """
        if indices is None:
            indices = list(range(scaffold.GetNumAtoms()))
        refined = Chem.RWMol(scaffold)
        refconf = refined.GetConformer()
        positions = defaultdict(list)  # coordinates
        equivalence = defaultdict(list)  # atom indices of hits.
        for h in self.hits:
            if h.GetProp('_Name') in self.unmatched:
                continue
            hc = h.GetConformer()
            for k, v in self.get_positional_mapping(scaffold, h).items():
                positions[k].append([hc.GetAtomPosition(v).x, hc.GetAtomPosition(v).y, hc.GetAtomPosition(v).z])
                equivalence[k].append(f'{h.GetProp("_Name")}.{v}')
        for i in range(scaffold.GetNumAtoms()):
            if i not in indices:
                continue
            elif len(positions[i]) == 0:
                refined.GetAtomWithIdx(i).SetDoubleProp('_Stdev', 0.)
                refined.GetAtomWithIdx(i).SetDoubleProp('_Max', 0.)
                refined.GetAtomWithIdx(i).SetProp('_Origin', 'none')
                # warn(f'Atom {i}  {scaffold.GetAtomWithIdx(i).GetSymbol}/{refined.GetAtomWithIdx(i).GetSymbol} '+ \
                #     'in scaffold that has no positions.')
            else:
                p = np.mean(np.array(positions[i]), axis=0).astype(float)
                # sd = np.mean(np.std(np.array(positions[i]), axis=0)).astype(float)
                ds = [np.linalg.norm(p - pi) for pi in positions[i]]
                sd = np.std(ds)
                md = np.max(ds)
                refined.GetAtomWithIdx(i).SetProp('_Origin', json.dumps(equivalence[i]))
                refined.GetAtomWithIdx(i).SetDoubleProp('_Stdev', sd)
                refined.GetAtomWithIdx(i).SetDoubleProp('_Max', md)
                if self.average_position:
                    refconf.SetAtomPosition(i, Point3D(p[0], p[1], p[2]))
        Chem.SanitizeMol(refined,
                         sanitizeOps=Chem.rdmolops.SanitizeFlags.SANITIZE_ADJUSTHS +
                                     Chem.rdmolops.SanitizeFlags.SANITIZE_SETAROMATICITY,
                         catchErrors=True)
        return refined

    # currently used only by partial
    # but maybe useful for other things
    def propagate_alternatives(self, fewer: List[Chem.Mol]) -> int:
        """
        Given the alt atoms strored in the Chem.Atom property ``_AltSymbol`` try those
        """
        pt = Chem.GetPeriodicTable()
        new = 0
        for template in list(fewer):
            for i, atom in enumerate(template.GetAtoms()):
                if atom.HasProp('_AltSymbol'):
                    alt = Chem.Mol(template)
                    aa = alt.GetAtomWithIdx(i)
                    aa.SetAtomicNum(pt.GetAtomicNumber(atom.GetProp('_AltSymbol')))
                    aa.ClearProp('_AltSymbol')
                    atom.ClearProp('_AltSymbol')
                    fewer.append(alt)
                    new += 1
        return new

    def pretweak(self) -> None:
        """
        What if the fragments were prealigned slightly? Really bad things happen.
        Nothing currently uses this without user interverntion.

        :return:
        """
        warn('This method is unreliable. Do not use it')
        ref = self.hits[0]
        for target in self.hits[1:]:
            A2B = list(self.get_positional_mapping(target, ref, 0.5).items())
            if A2B:
                rdMolAlign.AlignMol(target, ref, atomMap=A2B, maxIters=500)
            else:
                warn(f'No overlap? {A2B}')

    def sample_new_conformation(self, random_seed=None):
        """This method is intended for Multivictor.
        It generates a new conformation based on different random seeds.
        """
        scaffold = self.modifications["chimera"]
        atom_map = self.get_atom_map_fromProp(scaffold)
        if random_seed is None:
            random_seed = self.random_seed
        new_mol = self.place_from_map(target_mol=Chem.Mol(self.initial_mol), template_mol=scaffold, atom_map=atom_map,
                                      random_seed=random_seed)

        merging_mode = getattr(self, "merging_mode", "off")
        if merging_mode == 'off':
            pass
        elif merging_mode == 'full':
            pass
        elif merging_mode == 'partial':
            self.partial_blending()  # noqa will be in the subclasse partial
        elif merging_mode in ('none_permissive', 'permissive_none', 'none', 'expansion'):
            new_mol = self.posthoc_refine(new_mol)
        else:
            valid_modes = ('full', 'partial', 'none', 'none_permissive', 'expansion', 'off')
            raise ValueError(
                f"Merging mode can only be {'| '.join(valid_modes)}, not '{merging_mode}'")

        self.positioned_mol = new_mol
        return new_mol


    @classmethod
    def get_best_scoring(cls, mols: List[Chem.RWMol]) -> Chem.Mol:
        """
        Sorts molecules by how well they score w/ Merch FF
        """
        if len(mols) == 0:
            raise ValueError(f'No molecules')
        elif len(mols) == 1:
            return mols[0]
        # This is not expected to happen, but just in case
        mols = [m for m in mols if m is not None]
        scores = *map(cls.score_mol, mols),
        cls.journal.debug(f'`.get_best_scoring (unmerge)` Scores: {scores}')
        # proof that the mol has/lacks origins data:
        # for mol in mols:
        #     print('DEBUG OF LAST RESORT', mol)
        #     print(cls.origin_from_mol(cls, mol.GetMol())) # called from instance
        mol_scores = sorted(list(zip(mols, scores)), key=operator.itemgetter(1))
        return mol_scores[0][0]

    @staticmethod
    def score_mol(mol: Chem.Mol) -> float:
        """
        Scores a mol without minimising
        """
        if isinstance(mol, Chem.RWMol):
            mol = mol.GetMol()
        else:
            mol = Chem.Mol(mol)
        mol.UpdatePropertyCache()  # noqa
        Chem.SanitizeMol(mol)
        p = AllChem.MMFFGetMoleculeProperties(mol, 'MMFF94')
        if p is None:
            return float('nan')
        ff = AllChem.MMFFGetMoleculeForceField(mol, p)
        return ff.CalcEnergy()



from typing import Sequence
from rdkit import Chem
from rdkit.Chem import AllChem, Draw
from IPython.display import display
from unittest.mock import Mock
from typing import List

DISPLAYMODE = 'rdkit'
try:
    # Optional. See docstring in .ngl_display for more.
    from .ngl_display import MolNGLWidget, nv, ComponentViewer
    NGLWidget = nv.NGLWidget
    DISPLAYMODE = 'ngl'
except Exception as error:
    MolNGLWidget = Mock(name='MolNGLWidget')
    ComponentViewer = Mock(name='nglview.component.ComponentViewer')
    NGLWidget = Mock(name='nglview.NGLWidget')

try:
    from .mol3d_display import patched_3Dmol_view
    from .mol3d_display import monkey_patch as py3Dmol_monkey_patch
    # monkey_patch is a function that needs to be called but it is attached as a cls method to py3Dmol.view
    # which for clarity is rebranded as ``patched_3Dmol_view``
    DISPLAYMODE = 'py3Dmol'
except Exception as error:
    patched_3Dmol_view = Mock(name='py3Dmol.view')
    py3Dmol_monkey_patch = Mock(name='py3Dmol.view.monkey_patch')
from .branding import divergent_colors


def color_in(mols: List[Chem.Mol], color_scale=None, skip_feija=False):
    """
    assigns a color property to a mol based on color_scales of correct length

    In the `divergent_colors` first colour is the Fragmenstein colour (feijoa). Setting `color_in(False)` will skip it,
    allowing it to be used later on.
    """
    n_mols = len(mols)
    if n_mols == 0:
        return
    if color_scale is None and skip_feija:
        color_scale = divergent_colors[n_mols + int(skip_feija)][1:]
    elif color_scale is None:
        color_scale = divergent_colors[n_mols]
    elif len(color_scale) < n_mols:
        raise ValueError(f'color_scale is too short for {n_mols} mols.')
    else:
        pass
    for mol, color in zip(mols, color_scale):
        mol.SetProp('color', color)


def display_mols(mols: Sequence[Chem.Mol],
                 molsPerRow=5,
                 subImgSize=(150, 150),
                 useSVG=True):
    """
    Generic wrapper for calling ``display(Draw.MolsToGridImage)``
    """
    from matplotlib.colors import ColorConverter

    if not mols:
        return  # no mols, no display
    flattos = [AllChem.RemoveHs(mol) for mol in mols if isinstance(mol, Chem.Mol)]
    for mol in flattos:
        AllChem.Compute2DCoords(mol)
    dopts = Draw.rdMolDraw2D.MolDrawOptions()  # noqa
    dopts.addAtomIndices = True
    hex_map = {atom.GetIdx(): atom.GetProp('color') for atom in mol.GetAtoms()
               if atom.HasProp('color') and atom.GetProp('color')}
    rgb_map = {i: ColorConverter().to_rgb(n) for i, n in hex_map.items()}
    dopts.highlightAtomColors= list(rgb_map.values())
    dopts.highlightAtoms= list(rgb_map.keys())
    display(Draw.MolsToGridImage(flattos,
                                 legends=[mol.GetProp('_Name') if mol.HasProp('_Name') else '-' for mol in mols],
                                 subImgSize=subImgSize,
                                 useSVG=useSVG,
                                 molsPerRow=molsPerRow,
                                 drawOptions=dopts))




from rdkit import Chem
from rdkit.Chem import AllChem

from ._victor_base import _VictorBase
from ..m_rmsd import mRMSD
from typing import *


class _VictorValidate(_VictorBase):

    def validate(self, reference_mol: Chem.Mol) -> Dict[str, float]:
        """
        Get how well the results compare.
        Alternative, do a docking with victor.dock() (-> Chem.Mol)

        :param reference_mol: Crystal structure mol
        :return:
        """
        self.reference_mol : Chem.Mol = reference_mol
        return dict(minimized2hits_rmsd=self._get_minimized2hits_rmsd(),
                    reference2hits_rmsd=self._get_reference2hits_rmsd(),
                    reference2minimized_rmsd=self._get_reference2minimized_rmsd(),
                    )

    def _calc_rmsd_safeguard(self, func: Callable, *args, **kwargs):
        try:
            # compare with reference mol
            return func(*args, **kwargs).mrmsd
        except KeyboardInterrupt as err:
            raise err
        except self.error_to_catch as err:
            self.journal.error(f'{err.__class__.__name__}: {err} in validation step.')
            pass
        return float('nan')

    def _get_reference2hits_rmsd(self) -> float:
        return self._calc_rmsd_safeguard(mRMSD.from_other_annotated_mols,
                                         followup=self.reference_mol,
                                         hits=self.hits,
                                         annotated=self.monster.positioned_mol)

    def _get_reference2minimized_rmsd(self) -> float:
        return self._calc_rmsd_safeguard(mRMSD.from_unannotated_same_mols,
                                         moved_followup=self.minimized_mol,
                                         placed_followup=self.reference_mol)

    def _get_minimized2hits_rmsd(self) -> float:
        # actually is annotated, but best do the full loop in case it was tinkered with.
        return self._calc_rmsd_safeguard(mRMSD.from_other_annotated_mols,
                                         followup=self.minimized_mol,
                                         hits=self.hits,
                                         annotated=self.monster.positioned_mol)





# this is to keep mypy happy
from .pyrosetta_import import pyrosetta
import warnings
from typing import List, Tuple

class _IgorBase:
    def __init__(self):
        warnings.warn('Virtual only method: THIS METHOD SHOULD NOT BE RUN. CALL `Igor`', category=SyntaxWarning)
        self.pose = pyrosetta.Pose()
        self.constraint_file = ''
        self.ligand_residue: List[int] = []
        self.key_residues: List[int] = []
        self.atom_pair_constraint = 10
        self.angle_constraint = 10
        self.coordinate_constraint = 1
        self.fa_intra_rep = 0.005 # default

    def _vector2residues(self, vector: pyrosetta.Vector1) -> List[int]:
        """
        This method is for machines. See ``residues_in_vector`` instead.
        """
        return [i + 1 for i, v in enumerate(vector) if v == 1]


import unittest

# ======================================================================================================================
from rdkit import Chem
from rdkit.Chem import AllChem

from fragmenstein import Monster, Igor
from fragmenstein.mpro import MProVictor
from typing import Union
# ======================================================================================================================
class UnresolvedProblems(unittest.TestCase):

    def setUp(self):
        Igor.init_pyrosetta()

    def assertMolEqual(self, mol1: Union[str, Chem.Mol], mol2: Union[str, Chem.Mol]):
        """
        For now I am comparing SMILES (tut tut), but I will fix it one day...
        """
        smiles1: str = mol1 if isinstance(mol1, str) else Chem.MolToSmiles(AllChem.RemoveAllHs(mol1))
        smiles2: str = mol1 if isinstance(mol2, str) else Chem.MolToSmiles(AllChem.RemoveAllHs(mol2))
        self.assertEqual(smiles1, smiles2)

    def test_recto_fail_A(self):
        """This used to fail."""
        MProVictor.monster_throw_on_discard = True
        victor = MProVictor.from_hit_codes(hit_codes=['x11612', 'x11475'])
        victor.combine()
        self.assertEqual(victor.error_msg, '', victor.error_msg)

    def test_supplementary1_to_recto_fail_A(self):
        """
        This was meant to test the above, but it works fine.
        :return:
        """
        # make hits
        chlorotoluene = Chem.MolFromSmiles('c1c(Cl)c(C)ccc1')
        AllChem.EmbedMolecule(chlorotoluene)
        chlorotoluene.SetProp('_Name', 'orthochlorotoluene')
        toluene = Chem.RWMol(chlorotoluene)
        toluene.RemoveAtom(2)
        Chem.SanitizeMol(toluene)
        toluene.SetProp('_Name', 'toluene')
        chlorobutane = Chem.RWMol(chlorotoluene)
        for n in range(chlorobutane.GetNumAtoms() - 1, 4, -1):
            chlorobutane.RemoveAtom(n)
        for atom in chlorobutane.GetAtoms():
            atom.SetIsAromatic(False)
        for bond in chlorobutane.GetBonds():
            bond.SetBondType(Chem.BondType.SINGLE)
        Chem.SanitizeMol(chlorobutane)
        chlorobutane.SetProp('_Name', '2-chlorobutane')
        # merge
        monster = Monster(hits=[toluene, chlorobutane]).combine(keep_all=False)
        # ======
        self.assertMolEqual(chlorotoluene, monster.positioned_mol)

    def test_supplementary2_to_recto_fail_A(self):
        """
        This was meant to test as above.
        It merges xylene with chloropentane
        :return:
        """
        #
        chloroxylene = Chem.MolFromSmiles('Cc1c(Cl)c(C)ccc1')
        AllChem.EmbedMolecule(chloroxylene)
        chloroxylene.SetProp('_Name', 'chloroxylene')
        #
        xylene = Chem.RWMol(chloroxylene)
        xylene.RemoveAtom(3)
        Chem.SanitizeMol(xylene)
        xylene.SetProp('_Name', 'xylene')
        #
        chloropentane = Chem.RWMol(chloroxylene)
        for n in range(chloropentane.GetNumAtoms() - 1, 5, -1):
            chloropentane.RemoveAtom(n)
        for atom in chloropentane.GetAtoms():
            atom.SetIsAromatic(False)
        for bond in chloropentane.GetBonds():
            bond.SetBondType(Chem.BondType.SINGLE)
        Chem.SanitizeMol(chloropentane)
        chloropentane.SetProp('_Name', '2-chloropentane')
        #
        monster = Monster(hits=[xylene, chloropentane]).combine(keep_all=False)
        # ======
        self.assertEqual(Chem.MolToSmiles(chloroxylene), Chem.MolToSmiles(AllChem.RemoveAllHs(monster.positioned_mol)))

    def test_longer_link(self):
        """
        This is a know failure of the algorithm.
        It does not map "properly" the case where a link is lengthened. This is problematic in general
        and MCS based approach would not work even if the bondcomparison took into account whether atom is part
        of a linear stretch.
        """
        benzylbenzene = Chem.MolFromSmiles('c1ccccc1Cc1ccccc1')
        benzylbenzene.SetProp('_Name', 'benzylbenzene')
        assert AllChem.EmbedMolecule(benzylbenzene) == 0
        # extra carbon in bridge... dibenzyl
        monster = Monster([benzylbenzene, ]).place_smiles('c1ccccc1CCc1ccccc1')
        # this _seems_ correct, but it is not as the second ring is not mapped correctly.
        self.assertEqual(benzylbenzene.GetNumAtoms(),
                         len(monster.convert_origins_to_custom_map()['benzylbenzene']),
                         'not all atoms are mapped')

if __name__ == '__main__':
    unittest.main()



from ._combine import LabCombine
from ._place import LabPlace
from ._base import binarize, unbinarize
from ._place import MolPlacementInput, BinPlacementInput
from ._extras import LabExtras
from ._score import LabScore
try:
    from .validator import place_input_validator
except Exception as error:
    from warnings import warn
    warn(f'Supressed {error.__class__.__name__}: {error}')
    place_input_validator = None

class Laboratory(LabCombine, LabPlace, LabExtras, LabScore):
    """
    This class runs the combination or placement tasks of a list of molecules as subprocesses.
    The module used is ``pebble``, which is the same as ``multiprocessing`` but with a few more features.
    For more advanced usages, dask is a good candidate,
    but dask jobs on a cluster with a batch-queueing engine are non-trivial
    to set up.

    .. code-block:: python
        lab = Laboratory(pdbblock=template, covalent_resi=covalent_resi)
        results: pd.DataFrame = lab.combine(hits)
        px.histogram(results,
                     x='outcome',
                     category_orders={'outcome': lab.category_labels},
            title='Distribution of outcome')

    In the case of subclassed Victor, the user can create a `Laboratory` instance,
    and replace `.Victor` with their subclass of Victor.
    However, there are a few requirements, so please see ``combine_subprocess`` and ``place_subprocess``,
    where it gets called.
    Laboratory does not contain handy empty methods for the user to override like Victor (

    `post_igor_step`)
    """


import argparse
from .base import set_verbose
from pathlib import Path
from ..igor import Igor, pyrosetta
from collections import Counter

class FragmensteinParserUtils:

    def _define_utils(self, parser: argparse.ArgumentParser):
        subparsers = parser.add_subparsers(title='operation', help='minimize or ...')
        minimize_parser = subparsers.add_parser('minimize',
                                               help='prepare (minimise) the structure and get rid of waters etc')
        minimize_parser.set_defaults(func=self.igor_minimize)
        minimize_parser.add_argument('-t', '--template', required=True, help='Template PDB file')
        minimize_parser.add_argument('-o', '--output', default='', help='output PDB folder')
        minimize_parser.add_argument('-v', '--verbose', action="count", help='verbose')
        minimize_parser.add_argument('-ed', '--electron_density', default='', help='electron density map')
        minimize_parser.add_argument('-cw', '--constraint_weight', default=5, help='constraint weight')
        minimize_parser.add_argument('-c', '--cycles', default=15, help='number of cycles')
        minimize_parser.add_argument('-cf', '--constraint_file', default='', help='constraint file')
        minimize_parser.add_argument('-first', '--first_chain_only', default=False, help='only keep first chain')

    def igor_minimize(self, args: argparse.Namespace):
        set_verbose(args.verbose)
        Igor.init_pyrosetta()
        pyrosetta.rosetta.basic.options.set_boolean_option('run:ignore_zero_occupancy', False)
        pyrosetta.rosetta.basic.options.set_boolean_option('in:auto_setup_metals', True)
        # ## Prep
        assert Path(args.template).exists(), f'{args.template} does not exist'
        if args.output == '':
            args.output = args.template.replace('.pdb', '_min.pdb')
        assert not Path(args.outfile).exists(), f'{args.outfile} already exists'
        # ## Read pose
        pose: pyrosetta.Pose = pyrosetta.pose_from_file(args.template)
        con_tally = Counter(
            [type(con).__name__ for con in pose.constraint_set().get_all_constraints()]).most_common()
        print('chains', pyrosetta.rosetta.core.conformation.pose.conf2pdb_chain(pose))
        print('sequence', pose.sequence())
        print('Constraints present', con_tally)
        if args.first_chain_only:
            original: pyrosetta.Pose = pose.split_by_chain(1)
        if args.constraint_file:
            Igor.add_constraint_file(pose,
                                     constraint_file=args.constraint_file)
        # ## Relax
        if args.electron_density == '':
            relaxed: pyrosetta.Pose = Igor.relax_with_ED(pose,
                                                           ccp4_file=args.ccp4_file,
                                                           constraint_weights=(args.constraint_weight,),
                                                           cycles=args.cycles)
        else:
            relaxed: pyrosetta.Pose = Igor.tethered_relax(pose,
                                                          constraint_weight=args.constraint_weight,
                                                          cycles=args.cycles)
        # ## Dump
        relaxed.dump_pdb(args.outfile)



import json
import os

from rdkit import Chem

from ._victor_overridables import _VictorOverridables


class _VictorStore(_VictorOverridables):
    # _save_prerequisites is in VictorCommon

    def checkpoint(self):
        self._checkpoint_alpha()
        self._checkpoint_bravo()
        self._checkpoint_charlie()

    def _checkpoint_alpha(self):
        from ..igor.pyrosetta_import import pyrosetta
        self._log_warnings()
        # saving hits (without copying)
        for h, hit in enumerate(self.hits):
            if hit.HasProp("_Name") and hit.GetProp("_Name").strip():
                name = hit.GetProp("_Name")
            else:
                name = f'hit{h}'
            hfile = os.path.join(self.work_path, self.long_name, f'{name}.pdb')
            Chem.MolToPDBFile(hit, hfile)
            mfile = os.path.join(self.work_path, self.long_name, f'{name}.mol')
            Chem.MolToMolFile(hit, mfile, kekulize=False)
        # saving params template
        params_template_file = os.path.join(self.work_path, self.long_name, self.long_name + '.params_template.pdb')
        Chem.MolToPDBFile(self.params.mol, params_template_file)
        params_template_file = os.path.join(self.work_path, self.long_name, self.long_name + '.params_template.mol')
        Chem.MolToMolFile(self.params.mol, params_template_file)
        # checking all is in order
        ptest_file = os.path.join(self.work_path, self.long_name, self.long_name + '.params_test.pdb')
        self.unbound_pose.dump_pdb(ptest_file)
        pscore_file = os.path.join(self.work_path, self.long_name, self.long_name + '.params_test.score')
        scorefxn = pyrosetta.get_fa_scorefxn()
        with open(pscore_file, 'w') as w:
            w.write(str(scorefxn(self.unbound_pose)))
        self._log_warnings()

    def _checkpoint_bravo(self):
        self._log_warnings()
        self.journal.debug(f'{self.long_name} - saving mols from monster')
        # if self.monster.scaffold is not None:
        #     scaffold_file = os.path.join(self.work_path, self.long_name, self.long_name + '.scaffold.mol')
        #     Chem.MolToMolFile(self.monster.scaffold, scaffold_file, kekulize=False)
        #     if self.monster.scaffold.HasProp('parts'):
        #         disregard = json.loads(self.monster.scaffold.GetProp('parts'))
        #         self.journal.info(f'{self.long_name} - disregarded {disregard}')
        #     else:
        #         disregard = []
        # if self.monster.chimera is not None:
        #     chimera_file = os.path.join(self.work_path, self.long_name, self.long_name + '.chimera.mol')
        #     Chem.MolToMolFile(self.monster.chimera, chimera_file, kekulize=False)
        if self.monster.positioned_mol is not None:
            pos_file = os.path.join(self.work_path, self.long_name, self.long_name + '.positioned.mol')
            Chem.MolToMolFile(self.monster.positioned_mol, pos_file, kekulize=False)
        if self.monster.mol_options:
            opt_file = os.path.join(self.work_path, self.long_name, self.long_name + '.mol_options.sdf')
            writer = Chem.SDWriter(opt_file)
            writer.SetKekulize(False)
            for t in self.monster.mol_options:
                writer.write(t)
            writer.close()
        frag_file = os.path.join(self.work_path, self.long_name, self.long_name + '.monster.json')
        data = {'smiles': self.smiles,
                'origin': self.monster.origin_from_mol(self.monster.positioned_mol),
                'stdev': self.monster.stdev_from_mol(self.monster.positioned_mol)}
        # if disregard:
        #     data['disregard'] = disregard
        with open(frag_file, 'w') as w:
            json.dump(data, w)
        self._log_warnings()
        # unminimized_pdbblock will be saved by igor (round trip via pose)

    def _checkpoint_charlie(self):
        self._log_warnings()
        self.journal.debug(f'{self.long_name} - saving pose from igor')
        min_file = os.path.join(self.work_path, self.long_name, self.long_name + '.holo_minimised.pdb')
        self.igor.pose.dump_pdb(min_file)
        self.journal.debug(f'{self.long_name} - calculating Gibbs')
        # recover bonds
        lig_file = os.path.join(self.work_path, self.long_name, self.long_name + '.minimised.mol')
        Chem.MolToMolFile(self.minimized_mol, lig_file)
        score_file = os.path.join(self.work_path, self.long_name, self.long_name + '.minimised.json')
        with open(score_file, 'w') as w:
            json.dump({'Energy': self.energy_score,
                       'mRMSD': self.mrmsd.mrmsd,
                       'RMSDs': self.mrmsd.rmsds}, w)
        self._log_warnings()


from __future__ import annotations
from io import StringIO
from typing import Dict, TYPE_CHECKING, Tuple, Union, List, Optional, Any
import sys

from ..display import MolNGLWidget, DISPLAYMODE, py3Dmol_monkey_patch, patched_3Dmol_view  # real or a mock
from IPython.display import display
from rdkit import Chem
from rdkit.Chem import AllChem, Draw
from ._base import WaltonBase


class WaltonArt(WaltonBase):

    def _mol2svg(self, mol: Chem.Mol) -> str:  # noqa it is static, but I dont care.
        d2d = Draw.rdMolDraw2D.MolDraw2DSVG(250, 200)
        mol2 = AllChem.RemoveHs(mol)
        AllChem.Compute2DCoords(mol2)
        atoms = mol2.GetNumAtoms()
        for idx in range(atoms):
            mol2.GetAtomWithIdx(idx).SetProp('molAtomMapNumber', str(mol2.GetAtomWithIdx(idx).GetIdx()))
        d2d.DrawMolecule(mol2)
        d2d.FinishDrawing()  # noqa it is filled.
        return d2d.GetDrawingText()  # noqa it is filled.

    def _get_mol_name(self, mol: Chem.Mol) -> str:  # noqa
        if mol.HasProp('_Name'):
            return mol.GetProp('_Name')
        else:
            return '「 no name 」'

    def _get_mol_color(self, mol: Chem.Mol) -> str:  # noqa
        if mol.HasProp('color'):
            return mol.GetProp('color')
        else:
            return 'gray'

    def _mol2smiles(self, mol: Chem.Mol) -> str:  # noqa
        mol2 = AllChem.RemoveHs(mol)
        return Chem.MolToSmiles(mol2)

    def _get_mol_details(self, mol: Chem.Mol) -> Dict[str, str]:
        return dict(name=self._get_mol_name(mol),
                    smiles=self._mol2smiles(mol),
                    svg=self._mol2svg(mol),
                    color=self._get_mol_color(mol),
                    )

    def _repr_html_(self):
        inner = '<p style="color: {color};"><b>{name}</b><br/>{smiles}</p>{svg}'
        template = f'<div style="float: left; padding: 10px;">{inner}</div>'
        return "\n".join(template.format(**self._get_mol_details(mol)) for mol in self.mols + [self.merged] if mol)

    # ## 3D --------------------------------------------

    def to_nglview(self) -> MolNGLWidget:
        """`
        generates a NGLWidget (``IPython.display.display`` will show it)
        with the compounds and the merged if present.
        The colours will be those in the ``Chem.Mol``'s property ``_color``.
        """
        self.monster.journal('This function is deprecated, use show instead')
        view = MolNGLWidget()
        for mol in self.mols + [self.merged]:
            if not mol:  # self.merged is empty
                continue
            view.add_mol(mol)
        return view

    def refresh_nglview(self, view: MolNGLWidget, *others:Chem.Mol) -> None:
        """
        In Walton altering hits manually does not have an effect on the NGLWidget,
        but altering hits has no bad juju unlike Monster where altering the `.mols` can have a detrimental effect,
        due to the lack of the sanitisation from the `monster.fix_mols(mols)` call.

        :param view:
        :return:
        """
        view.remove_all_components()
        for mol in self.mols + [self.merged] + list(others):
            if not mol:  # self.merged is empty
                continue
            view.add_mol(mol)


    def show3d(self) -> None:
        """
        Shows the structures both in 2d and 3d. Which is weird. Hence the deprecation.
        """
        self.monster.journal.warn('show3d is deprecated, use show instead')
        display(self)
        self.show()
        
    def show(self, to_display: bool = False, show_positioned_mol:bool=True, viewer_mode=DISPLAYMODE) \
                                                                -> Tuple[Union[MolNGLWidget,patched_3Dmol_view], str]:
        """
        Shows the structures in 3d and behaves like ``Monster.show``
        display(self)  will show in 2d molecules...
        """
        return self.monster.show(to_display=to_display, show_positioned_mol=show_positioned_mol, viewer_mode=viewer_mode)



from ._utility import _MonsterUtil
from rdkit import Chem
from rdkit.Chem import AllChem, rdqueries, rdMolAlign
from rdkit import ForceField as FF
from typing import Optional, List, Union, Tuple, Dict
from warnings import warn
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt
from ..error import FragmensteinError


@dataclass
class MinizationOutcome:
    success: bool
    mol: Chem.Mol
    ideal: Chem.Mol
    U_pre: float = float('nan')
    U_post: float = float('nan')
    delta: float = float('nan')


class _MonsterFF(_MonsterUtil):

    def mmff_minimize(self,
                      mol: Optional[Chem.Mol] = None,
                      neighborhood: Union[Chem.Mol, None] = None,
                      ff_max_displacement: float = 0.,
                      ff_constraint: int = 10,
                      ff_max_iterations: int=200,
                      ff_cutoff: float = 100.,
                      allow_lax: bool = True) -> MinizationOutcome:
        """
        Minimises a mol, or self.positioned_mol if not provided, with MMFF constrained to ff_max_displacement Å.
        Gets called by Victor if the flag .monster_mmff_minimisation is true during PDB template construction.

        :param mol: Molecule to minimise. If None, self.positioned_mol is used.
        :param neighborhood: Protein neighboorhood (ignored if None)
        :param ff_max_displacement: Distance threshold (Å) for atomic positions mapped to hits for  MMFF constrains.
                            if NaN then fixed point constraints (no movement) are used.
                            This is passed as maxDispl to MMFFAddPositionConstraint.
        :param ff_constraint: Force constant for MMFF constraints.
        :param ff_cutoff: kcal/mol diff value to consider a failed minimisation.
        :param allow_lax: If True and the minimisation fails, the constraints are halved and the minimisation is rerun.
        :return: None

        Note that most methods calling this via Victor
        now use its ``.settings.py['ff_max_displacement']`` and ``.settings.py['ff_constraint']``
        and do not use the defaults.
        """
        # ## input fixes
        if mol is None and self.positioned_mol is None:
            raise ValueError('No valid molecule')
        elif mol is None:
            mol = self.positioned_mol
        else:
            pass  # mol is fine
        # ## prep
        success: bool
        fixed_mode = str(ff_max_displacement).lower() == 'nan'
        mol = AllChem.AddHs(mol, addCoords=True)
        # protect
        for atom in mol.GetAtomsMatchingQuery(Chem.rdqueries.HasPropQueryAtom('_IsDummy')):
            atom.SetAtomicNum(8)
        combo, fixed_idxs = self._prep_combined(mol, neighborhood)
        ideal: Chem.Mol = self.make_ideal_mol(mol, ff_minimise=True)
        ideal_E: float = ideal.GetDoubleProp('Energy')
        # ## Start FF
        p = AllChem.MMFFGetMoleculeProperties(combo, 'MMFF94')
        if p is None:
            self.journal.error(f'MMFF cannot work on a molecule that has errors!')
            return MinizationOutcome(success=False, mol=mol, ideal=ideal)
        ff = AllChem.MMFFGetMoleculeForceField(combo, p, ignoreInterfragInteractions=False)
        if ff is None:
            return MinizationOutcome(success=False, mol=mol, ideal=ideal,
                                     U_post=float('nan'), U_pre=float('nan'), delta=0)
        # restrain
        restrained = []
        # mol not combo here:
        conserved: List[Chem.Atom] = list(mol.GetAtomsMatchingQuery(rdqueries.HasPropQueryAtom('_Novel', negate=True)))
        # weird corner case
        if len(conserved) == mol.GetNumAtoms() and fixed_mode:
            ff.Initialize()
            dU: float = ff.CalcEnergy()
            self.journal.warning('No novel atoms found in fixed_mode (ff_max_displacement == NaN), ' + \
                                 'this is probably a mistake')
            # nothing to do...
            return MinizationOutcome(success=True, mol=mol, ideal=ideal, U_post=dU, U_pre=dU, delta=0)
        # constrain or freeze
        for atom in conserved:
            i = atom.GetIdx()
            if atom.GetAtomicNum() == 1:
                # let hydrogens move
                continue
            elif fixed_mode:
                ff.AddFixedPoint(i)
            else:
                # https://github.com/rdkit/rdkit/blob/115317f43e3bdfd73673ca0e4c6b4035aa26a034/Code/ForceField/UFF/PositionConstraint.cpp#L35
                ff.MMFFAddPositionConstraint(i, maxDispl=ff_max_displacement, forceConstant=ff_constraint)
            restrained.append(i)
        # constrain dummy atoms
        for atom in mol.GetAtomsMatchingQuery(rdqueries.HasPropQueryAtom('_IsDummy')):
            i = atom.GetIdx()
            ff.MMFFAddPositionConstraint(i, maxDispl=0, forceConstant=ff_constraint * 5)
            restrained.append(i)
        for i in fixed_idxs:  # neighborhood is frozen
            ff.AddFixedPoint(i)
        # ## Minimize
        try:
            dG_pre = ff.CalcEnergy()
            dG_post = dG_pre
            previous_dG = 0.
            m = -1
            # this is a bit of a hack, but it works to make sure its not a flipped plateau-like local minima
            while previous_dG == 0. or previous_dG - dG_post > 0.5:
                previous_dG = dG_post
                m = ff.Minimize(maxIts=ff_max_iterations)
                dG_post = ff.CalcEnergy()
                if m == -1:
                    break
            if m == -1:
                self.journal.error('MMFF Minisation could not be started')
                success = False
            elif m == 0:
                self.journal.info('MMFF Minisation was successful')
                success = True
            elif m == 1:
                self.journal.info('MMFF Minisation was run, but the minimisation was not unsuccessful')
                success = False
            else:
                self.journal.critical("Iä! Iä! Cthulhu fhtagn! Ph'nglui mglw'nafh Cthulhu R'lyeh wgah'nagl fhtagn")
                success = False
        except RuntimeError as error:
            self.journal.info(f'MMFF minimisation failed {error.__class__.__name__}: {error}')
            return MinizationOutcome(success=False, mol=mol, ideal=ideal)
        # extract
        new_mol = self.extract_from_neighborhood(combo)
        ligand_E: float = self.MMFF_score(new_mol, delta=False)
        new_mol.SetDoubleProp('Energy', ligand_E)
        # check
        if ligand_E - ideal_E > abs(ff_cutoff):
            success = False  # damn
        if not success and allow_lax:
            self.journal.debug(f'MMFF minimisation failed, trying again with lax constraint {ff_constraint // 5}')
            return self.mmff_minimize(mol,
                                      neighborhood=neighborhood,
                                      ff_max_displacement=ff_max_displacement,
                                      ff_constraint=ff_constraint // 5,
                                      allow_lax=False)
        # deprotect
        for atom in new_mol.GetAtomsMatchingQuery(Chem.rdqueries.HasPropQueryAtom('_IsDummy')):
            atom.SetAtomicNum(0)
        # prevent drift:
        #rdMolAlign.AlignMol(new_mol, mol, atomMap=list(zip(restrained, restrained)))
        self.journal.info(f'MMFF minimisation: {dG_pre:.2f} -> {dG_post:.2f} kcal/mol ' +
                           f'w/ {rdMolAlign.CalcRMS(new_mol, mol)}Å RMSD at ' +
                           f'max displacement={ff_max_displacement} & constraint={ff_constraint}'
                           )
        return MinizationOutcome(success=success,
                                 mol=new_mol,
                                 ideal=ideal,
                                 U_post=dG_post,
                                 U_pre=dG_pre,
                                 delta=dG_post - dG_pre)

    def _prep_combined(self, mol, neighborhood) -> Tuple[Chem.Mol, List[int]]:
        # ## protect (DummyMasker could be used here)
        for atom in mol.GetAtomsMatchingQuery(Chem.rdqueries.AtomNumEqualsQueryAtom(0)):
            atom.SetBoolProp('_IsDummy', True)
            atom.SetAtomicNum(16)
        Chem.SanitizeMol(mol)
        # ## Combine with neighborhood
        if neighborhood is not None:
            Chem.SanitizeMol(neighborhood)
            hydroneighborhood = AllChem.AddHs(neighborhood, addCoords=True)
            combo: Chem.Mol = Chem.CombineMols(mol, hydroneighborhood)
            Chem.SanitizeMol(combo)
            fixed_idxs: List[int] = list(range(mol.GetNumAtoms(), combo.GetNumAtoms()))
        else:
            combo = Chem.Mol(mol)
            fixed_idxs: List[int] = []
        self.journal.debug(f'Combined molecule (ligand+neighbourhood) has {combo.GetNumAtoms()} atoms, {fixed_idxs} fixed')
        return combo, fixed_idxs

    def MMFF_score(self, mol: Optional[Chem.Mol] = None, delta: bool = False, mode: str = 'MMFF') -> float:
        """
        Merck force field. Chosen over Universal for no reason at all.

        :param mol: ligand
        :type mol: ``Chem.Mol`` optional. If absent extracts from pose.
        :param delta: report difference from unbound (minimized)
        :type delta: bool
        :param mode: 'MMFF' or 'UFF'
        :type mode: str
        :return: kcal/mol
        :rtype: float

        :warning: This was moved out of Igor. Victor has the method for calling it with igor.mol_from_pose
        """
        if mol is None:
            mol = self.positioned_mol
        try:
            mol = AllChem.AddHs(mol)  # copy!
            if mode == 'UFF':
                ff = AllChem.UFFGetMoleculeForceField(mol)
            elif mode == 'MMFF':
                p = AllChem.MMFFGetMoleculeProperties(mol, 'MMFF94')
                ff = AllChem.MMFFGetMoleculeForceField(mol, p)
            else:
                raise ValueError(f'Unknown mode: {mode} (choice: MMFF or UFF)')
            ff.Initialize()
            # print(f'MMFF: {ff.CalcEnergy()} kcal/mol')
            if delta:
                pre = ff.CalcEnergy()
                ff.Minimize()
                post = ff.CalcEnergy()
                return pre - post
            else:
                return ff.CalcEnergy()
        except RuntimeError as err:
            self.journal.warning(f'{err.__class__.__name__}: {err} (It is generally due to bad sanitisation)')
            return float('nan')

    @classmethod
    def get_close_indices(cls, query: Chem.Mol, target: Chem.Mol, cutoff: float = 5.) -> List[int]:
        """
        Give an rdkit Chem.Mol ``query`` get the atom idices of ``target`` that are with ``cutoff`` Å.
        """
        combo = Chem.CombineMols(target, query)
        distances: npt.NDArray[np.float64] = AllChem.Get3DDistanceMatrix(combo)
        query2target_dist: npt.NDArray[np.float64] = distances[
            slice(target.GetNumAtoms(), combo.GetNumAtoms(), 1), slice(0, target.GetNumAtoms(), 1)].min(axis=0)
        neighbors: npt.NDArray[np.int64] = np.where(query2target_dist <= cutoff)[0]
        return list(map(int, neighbors))

    @classmethod
    def _get_aromatic_neighbors(cls, atom, accounted):
        neighbor: Chem.Atom
        for neighbor in atom.GetNeighbors():
            neigh_idx: int = neighbor.GetIdx()
            if neighbor.GetIsAromatic() and neigh_idx not in accounted:
                accounted.append(neigh_idx)
                return cls._get_aromatic_neighbors(neighbor, accounted)
        return accounted

    @classmethod
    def extract_atoms(cls, protein: Chem.Mol, keepers: List[int], expand_aromatics: bool = True) -> Chem.Mol:
        """
        Extract the given atom indices (``keepers``) from ``protein``.
        Expanding to full aromatic ring and copying conformers
        """
        pasteboard = Chem.RWMol()
        # ## Expand aromatic rings
        exkeepers = list(keepers)
        if expand_aromatics:
            for idx in keepers:
                atom: Chem.Atom = protein.GetAtomWithIdx(idx)
                if not atom.GetIsAromatic():
                    continue
                exkeepers = cls._get_aromatic_neighbors(atom, exkeepers)
        # ## Add atoms
        prot2paste: Dict[int, int] = {}
        for idx in exkeepers:
            atom: Chem.Atom = protein.GetAtomWithIdx(int(idx))  # no to np.int64
            if not expand_aromatics:
                atom.SetIsAromatic(False)
            prot2paste[idx] = pasteboard.AddAtom(atom)
        # ## Add bonds
        for prot_idx, paste_idx in prot2paste.items():
            atom: Chem.Atom = protein.GetAtomWithIdx(prot_idx)
            for prot_neighbor_idx in [n.GetIdx() for n in atom.GetNeighbors()]:
                if prot_neighbor_idx in prot2paste and prot_neighbor_idx > prot_idx:
                    prot_bond: Chem.Bond = protein.GetBondBetweenAtoms(prot_idx, prot_neighbor_idx)
                    paste_neighneighbor_idx = prot2paste[prot_neighbor_idx]
                    pasteboard.AddBond(paste_idx, paste_neighneighbor_idx,
                                       prot_bond.GetBondType() if expand_aromatics else Chem.BondType.SINGLE)

        pasteboard_conf = Chem.Conformer(len(keepers))
        positions: npt.NDArray = protein.GetConformer().GetPositions()
        for prot_idx, paste_idx in prot2paste.items():
            pasteboard_conf.SetAtomPosition(paste_idx, positions[prot_idx, :])
        pasteboard.AddConformer(pasteboard_conf)
        return pasteboard.GetMol()

    def get_neighborhood(self, apo_block: str, cutoff: float, mol: Optional[Chem.Mol] = None, addHs=True) -> Chem.Mol:
        """
        Get the neighborhood of the protein from the apo_block around the cutoff of the mol.
        Note: The atoms will have a prop ``IsNeighborhood`` which is used after it is combined.
        """
        if mol is None:
            mol = self.positioned_mol
        protein: Chem.Mol = Chem.MolFromPDBBlock(apo_block)
        neighbor_idxs: List[int] = self.get_close_indices(mol, protein, cutoff)
        neighborhood: Chem.Mol = self.extract_atoms(protein, neighbor_idxs)
        AllChem.SanitizeMol(neighborhood, catchErrors=True)
        if addHs:
            neighborhood = AllChem.AddHs(neighborhood, addCoords=True)
        self.journal.debug(f'{cutoff}Å Neighborhood has {neighborhood.GetNumAtoms()} atoms')
        for atom in neighborhood.GetAtoms():
            atom.SetBoolProp('IsNeighborhood', True)
        AllChem.SanitizeMol(neighborhood, catchErrors=True)
        return neighborhood

    def make_ideal_mol(self, mol: Optional[Chem.Mol]=None, ff_minimise: bool=False) -> Chem.Mol:
        if mol is None:
            mol = self.positioned_mol
        ideal = Chem.Mol(mol)
        ideal.SetDoubleProp('Energy', float('nan'))
        AllChem.EmbedMolecule(ideal)
        p: FF.MMFFMolProperties = AllChem.MMFFGetMoleculeProperties(ideal, 'MMFF94')
        ff = AllChem.MMFFGetMoleculeForceField(ideal, p)
        if ff is None:
            raise FragmensteinError('Ideal compound failed. Something is wrong with the SMILES')
        ff.Initialize()
        if ff_minimise:
            ff.Minimize()
        energy = ff.CalcEnergy()
        ideal.SetDoubleProp('Energy', energy)
        return ideal

    def extract_from_neighborhood(self, system: Chem.Mol) -> Chem.Mol:
        """
        Given a system of a neighbourhood + ligand extract everything that is not marked ``IsNeighborhood``.
        """
        rwmol = Chem.RWMol(system)
        rwmol.BeginBatchEdit()
        for atom in rwmol.GetAtoms():
            if atom.HasProp('IsNeighborhood'):
                rwmol.RemoveAtom(atom.GetIdx())
        rwmol.CommitBatchEdit()
        # isNeighborhood encompasses Hs... but just in case:
        # this warning happened: 'WARNING: not removing hydrogen atom without neighbors'
        # I have not seen it since, but I got a report of it
        new_mol = sorted(Chem.GetMolFrags(rwmol.GetMol(), asMols=True), key=Chem.Mol.GetNumAtoms, reverse=True)[0]
        return new_mol



import pandas as pd
from rdkit import Chem, rdBase, DataStructs
from rdkit.Chem import PandasTools
from .validator import place_input_validator
from .._cli_defaults import cli_default_settings
import time
from typing import List, Dict, Any, Optional


class LabExtras:
    error_classifications = {'SUCCESS': 'success',
                             'UNCLASSIFIED': 'unclassified',
                             'ConnectionError': 'distance',  # legacy
                             'DistanceError': 'distance',
                             'RectificationError': 'incorrect rectification',
                             'FullOverlapError': 'distance',
                             'TimeoutError': 'timeout',
                             'No sub-structure match': 'incorrect parameterisation #1',
                             'Violation occurred': 'incorrect rectification #1',
                             'KekulizeException': 'incorrect rectification #1',
                             'AtomValenceException': 'incorrect rectification #1',
                             'Bad Conformer Id': 'incorrect rectification #1',
                             ('rosetta', 'not found'): 'incorrect parameterisation #2',
                             ('rosetta', 'nan'): 'incorrect parameterisation #3',
                             'but not the minimised': 'incorrect parameterisation #4',
                             'UtilityExitException': 'incorrect parameterisation #5',
                             'TypeError': 'embedding error',
                             'RecursionError': 'embedding error',
                             'utterly impossible': 'embedding error'
                             }

    @classmethod
    def error_classify(cls, text):
        if not text:
            return cls.error_classifications['SUCCESS']
        text = str(text)
        for key, value in cls.error_classifications.items():
            if isinstance(key, tuple):
                if all(x in text for x in key):
                    return value
            elif key in text:
                return value
            else:
                return cls.error_classifications['UNCLASSIFIED']

    # these are properties of the dataframe that will be kept if present when saving to sdf
    default_sdf_properties = ['regarded', 'smiles', '∆∆G', '∆G_bound', '∆G_unbound',
                                         'comRMSD', 'N_constrained_atoms', 'N_unconstrained_atoms', 'runtime',
                                         'LE', 'outcome',
                                         'percent_hybrid']
    @classmethod
    def convert_to_sdf(self,
                       df: pd.DataFrame,
                       filename: str = f'fragmenstein.sdf',
                       acceptable_only=True,
                       sort_values: str = cli_default_settings['ranking'],
                       name_col: str = 'name',
                       mol_col='minimized_mol'):
        if acceptable_only:
            df = df.loc[(df.outcome == 'acceptable')]
        short = df.sort_values(sort_values).reset_index().drop_duplicates(name_col)
        valid_keys: List[str] = [k for k in self.default_sdf_properties if k in short.columns]
        PandasTools.WriteSDF(df=short,
                             out=filename,
                             molColName=mol_col,
                             idName=name_col,
                             properties=valid_keys
                             )

    # ---------- CLI --------------------------------------------

    @classmethod
    def core_ops(cls, hit_replacements, sw_databases, **settings):
        combinations: pd.DataFrame = cls._combine_ops(**settings)
        cls.correct_weaklings(hit_replacements, combinations)
        uncat_analogs: List[pd.DataFrame] = []
        for sw_db in sw_databases:
            s = cls.sw_search(combinations, sw_db=sw_db, **settings)
            s = s.loc[~s.hits.isna()]
            if len(s):
                uncat_analogs.append(s)
        assert uncat_analogs, 'No analogues were found!'
        analogs: pd.DataFrame = pd.concat(uncat_analogs, ignore_index=True).drop_duplicates('smiles')
        analogs.to_pickle(f'fragmenstein_analogs{settings["suffix"]}.pkl.gz')
        placements: pd.DataFrame = cls._place_ops(analogs=analogs, **settings)
        return placements

    @classmethod
    def _combine_ops(cls,
                     hits,
                     pdbblock,
                     suffix: str = cli_default_settings['suffix'],
                     n_cores: int = cli_default_settings['n_cores'],
                     combination_size: int = cli_default_settings['combination_size'],
                     timeout: int = cli_default_settings['timeout'],
                     max_tasks: int = cli_default_settings['max_tasks'],
                     blacklist: List[str] = cli_default_settings['blacklist'],
                     **settings) -> pd.DataFrame:
        """
        One of the operations of ``core_ops``.
        A thin wrapper around ``combine``.
        """
        lab = cls(pdbblock=pdbblock, covalent_resi=None)  # noqa it's inherited later
        lab.blacklist = blacklist
        tick = time.time()
        combinations: pd.DataFrame = lab.combine(hits,  # noqa it's inherited later
                                                 n_cores=n_cores,
                                                 timeout=timeout,
                                                 combination_size=combination_size,
                                                 max_tasks=max_tasks)
        combinations.to_pickle(f'fragmenstein_mergers{suffix}.pkl.gz')
        combinations.to_csv(f'fragmenstein_mergers{suffix}.csv')
        cls.Victor.journal.info(f'Combination {time.time() - tick} s')
        return combinations

    @classmethod
    def sw_search(cls, combinations: pd.DataFrame, suffix: str,
                  sw_dist: int, sw_length: int,
                  sw_db: str,
                  top_mergers: int=1_000,
                  ranking: Optional[str] = None, ranking_ascending: Optional[bool] = None,
                  sws: Optional = None,  # override smallworld api settings...
                  **setting) -> pd.DataFrame:
        """
        One of the operations of ``core_ops``.
        A thin wrapper around ``SmallWorld().search_many``.
        It will search for analogues in SmallWorld.
        To not go overboard ``top_mergers`` should probably not be greater than 1000 as a 1,000 calls is abusing it.
        If ranking is None then it assumed sorted.
        """
        if ranking is None:
            queries = combinations
        elif ranking_ascending is None:
            ranking_ascending = False if ranking in ('LE', 'N_interactions') else True
            queries = combinations.sort_values(ranking, ascending=ranking_ascending)
        else:
            queries = combinations.sort_values(ranking, ascending=ranking_ascending)
        smiles_col: str
        for smiles_col in ['simple_smiles', 'SMILES', 'smiles', 'smile']:
            if smiles_col in combinations.columns:
                break
        else:
            raise ValueError(f'No smiles column in {combinations.columns}')
        queries.loc[(combinations.outcome == 'acceptable')] \
                .drop_duplicates(smiles_col) \
                .reset_index() \
                .head(top_mergers)
        if sws is None:
            from smallworld_api import SmallWorld, NoMatchError
            sws = SmallWorld()
        try:
            analogs = sws.search_many(queries[smiles_col].to_list(),
                                      dist=sw_dist,
                                      length=sw_length,
                                      db=sw_db,
                                      tolerated_exceptions=Exception)
            analogs.to_pickle(f'fragmenstein_analogues{suffix}.{sw_db}.pkl.gz')
            invalid = analogs.qrySmiles.astype(str).isin(['None', 'nan', ''])
            analogs = analogs.loc[~invalid]
        except NoMatchError as error:
            print(f'Found no analogues. Consider changing `sw_dist` (distance by N of mismatches)')
            return pd.DataFrame()
        print(f'Found {len(analogs)} analogues')
        # query_index was added clientside to keep track!
        analogs['catalogue'] = sw_db
        analogs['query_name'] = analogs.query_index.map(queries.name.to_dict())
        analogs['hits'] = analogs.query_index.map(queries.hit_mols.to_dict())
        analogs = analogs.loc[~analogs.hits.isna()]  # not sure how this is possible, but can happen?
        analogs['hit_names'] = analogs.hits.apply(lambda m: [mm.GetProp('_Name') for mm in m])
        analogs['minimized_merger'] = analogs.query_index.map(queries.minimized_mol.to_dict())
        analogs['unminimized_merger'] = analogs.query_index.map(queries.unminimized_mol.to_dict())
        analogs['name'] = analogs['id'] + ':' + analogs['query_name']
        analogs['smiles'] = analogs.hitSmiles.str.split(expand=True)[0]
        analogs['custom_map'] = analogs.apply(cls.get_custom_map, axis=1)
        analogs.to_pickle(f'fragmenstein_analogues{suffix}.{sw_db}.pkl.gz')
        return analogs

    @classmethod
    def _place_ops(cls, analogs, pdbblock, n_cores, timeout, suffix, **settings) -> pd.DataFrame:
        """
        This is the classmethod called by ``core_ops``.
        The instance method ``place`` does the actual work, this is a thin wrapper.
        """
        lab = cls(pdbblock=pdbblock, covalent_resi=None, run_plip=True)  # noqa it's inherited later
        placements: pd.DataFrame = lab.place(place_input_validator(analogs),
                                             n_cores=n_cores,
                                             timeout=timeout)
        placements.to_pickle(f'fragmenstein_placed{suffix}.pkl.gz')
        placements.to_csv(f'fragmenstein_placed{suffix}.csv')
        # print(placements.outcome.value_counts())
        return placements

    @classmethod
    def get_custom_map(cls, row: pd.Series) -> Dict[str, Dict[int, int]]:
        """
        SmallWorld returns a mapping of the indices.
        This functions maps the indices back to the original hit molecules.

        :param row: SW pd.Series (or dict)
        :return:
        """
        try:
            # faux PDB block to trick the safeguards against bad PDB blocks/filenames
            if hasattr(cls, 'settings'):
                temp = cls.Victor(row.hits, pdb_block=Chem.MolToPDBBlock(Chem.MolFromFASTA('A')), **cls.settings)
            else:
                temp = cls.Victor(row.hits, pdb_block=Chem.MolToPDBBlock(Chem.MolFromFASTA('A')))
            temp.monster.positioned_mol = row.unminimized_merger
            temp.minimized_mol = row.minimized_merger
            return temp.migrate_sw_origins(row)
        except Exception as error:
            cls.Victor.journal.warn(f'{error.__class__.__name__} {error}')
            return {}

    @classmethod
    def correct_weaklings(cls, hit_replacements: pd.DataFrame, target_df: pd.DataFrame):
        """
        Some hits are weaker than the original fragment.
        If compounds in ``target_df`` are weaker than the inspiration in ``hit_replacements``,
        then the outcome is changed to ``weaker``.
        """
        # just in case...
        if 'hit_mols' not in target_df.columns:
            return
        target_df['hit_names'] = target_df.hit_mols \
            .apply(lambda v: v if isinstance(v, list) else []) \
            .apply(lambda v: [m.GetProp('_Name') for m in v])
        # assign by name
        dG_mapping = hit_replacements.set_index('name')['∆∆G'].to_dict()
        dG_mapping.update({k.replace('-', '_'): v for k, v in dG_mapping.items()})
        get_lowest = lambda names: min([0] + [dG_mapping.get(name, 0) for name in names])
        target_df['lowest_hit_∆∆G'] = target_df.hit_names.apply(get_lowest)
        worseness_mask = (target_df['∆∆G'] > target_df['lowest_hit_∆∆G'] * 0.8) & (
                target_df.outcome == 'acceptable')
        target_df.loc[worseness_mask, 'outcome'] = 'weaker'

    @classmethod
    def replace_hits(cls,
                     pdbblock:str,
                     hits: List[Chem.Mol],
                     n_cores=1,
                     timeout=600,
                     suffix: str = '',
                     run_plip: bool = True,
                     **settings):
        """
        Redock, but place => replace.
        Id est score the hits.

        This is not called by ``core_ops``.
        """
        # place themselves for a ∆∆G score
        lab = cls(pdbblock=pdbblock, covalent_resi=None, run_plip=run_plip)
        selfies = pd.DataFrame([dict(name=hit.GetProp('_Name') + '_replaced',
                                     hits=[hit],
                                     smiles=Chem.MolToSmiles(hit)
                                     ) for hit in hits])
        replacements: pd.DataFrame = lab.place(place_input_validator(selfies), n_cores=n_cores, timeout=timeout)
        cls.fix_intxns(replacements)
        replacements['bleached_name'] = replacements['name']
        replacements['name'] = replacements.hit_mols.apply(lambda hits: hits[0].GetProp('_Name') if hits else '')
        replacements.to_pickle(f'fragmenstein_hit_replacements{suffix}.pkl.gz')
        return replacements



########################################################################################################################

__doc__ = \
    """
This is the ring collapsing code.

In Pymol, a ring diameter is 2.7894375324249268 Å.

    fab F
    print cmd.distance('/obj01///PHE`1/CE1','/obj01///PHE`1/CD2')
    """

########################################################################################################################

import itertools
import json
from collections import defaultdict
from functools import partial
from typing import Optional, Dict, List, Any, Tuple, Union, Callable

import numpy as np
from rdkit import Chem
from rdkit.Geometry.rdGeometry import Point3D

from ._join_neighboring import _MonsterJoinNeigh
from .bond_provenance import BondProvenance
from ..error import DistanceError, FragmensteinError


########################################################################################################################

class _MonsterRing(_MonsterJoinNeigh):

    def collapse_mols(self, mols: List[Chem.Mol]):
        mols = [self.collapse_ring(mol) for mol in mols]
        [self.offset(mol) for mol in mols]
        return mols

    # =========== Collapse & Expand ====================================================================================

    def collapse_ring(self, mol: Chem.Mol) -> Chem.Mol:
        """
        Collapses a ring(s) into a single dummy atom(s).
        Stores data as JSON in the atom.

        :param mol:
        :return:
        """
        self.store_positions(mol)
        mol = Chem.RWMol(mol)
        conf = mol.GetConformer()  # noqa
        center_idxs = []
        morituri = []
        old2center = defaultdict(list)
        # Store simple info
        for atomset in mol.GetRingInfo().AtomRings():
            morituri.extend(atomset)
            neighs = []
            neighbonds = []
            bonds = []
            xs = []
            ys = []
            zs = []
            elements = []
            # add elemental ring
            c = mol.AddAtom(Chem.Atom('C'))
            center_idxs.append(c)
            central = mol.GetAtomWithIdx(c)
            name = mol.GetProp('_Name') if mol.HasProp('_Name') else '???'
            central.SetProp('_ori_name', name),
            # get data for storage
            for i in atomset:
                old2center[i].append(c)
                atom = mol.GetAtomWithIdx(i)
                neigh_i = [a.GetIdx() for a in atom.GetNeighbors()]
                neighs.append(neigh_i)
                bond = [mol.GetBondBetweenAtoms(i, j).GetBondType().name for j in neigh_i]
                bonds.append(bond)
                pos = conf.GetAtomPosition(i)
                xs.append(pos.x)
                ys.append(pos.y)
                zs.append(pos.z)
                elements.append(atom.GetSymbol())
            # store data in elemental ring
            central.SetIntProp('_ori_i', -1)
            central.SetProp('_ori_is', json.dumps(atomset))
            central.SetProp('_neighbors', json.dumps(neighs))
            central.SetProp('_xs', json.dumps(xs))
            central.SetProp('_ys', json.dumps(ys))
            central.SetProp('_zs', json.dumps(zs))
            central.SetProp('_elements', json.dumps(elements))
            central.SetProp('_bonds', json.dumps(bonds))
            central.SetIsotope(len(atomset))
            conf.SetAtomPosition(c, Point3D(*[sum(axis) / len(axis) for axis in (xs, ys, zs)]))
        # Store complex info
        for atomset, center_i in zip(mol.GetRingInfo().AtomRings(), center_idxs):
            # bond to elemental ring
            central = mol.GetAtomWithIdx(center_i)
            neighss = json.loads(central.GetProp('_neighbors'))
            bondss = json.loads(central.GetProp('_bonds'))
            for neighs, bonds in zip(neighss, bondss):
                for neigh, bond in zip(neighs, bonds):
                    if neigh not in atomset:
                        bt = getattr(Chem.BondType, bond)
                        if neigh not in morituri:
                            mol.AddBond(center_i, neigh, bt)
                            new_bond = mol.GetBondBetweenAtoms(center_i, neigh)
                            BondProvenance.set_bond(new_bond, 'original')
                        else:
                            for other_center_i in old2center[neigh]:
                                if center_i != other_center_i:
                                    if not mol.GetBondBetweenAtoms(center_i, other_center_i):
                                        mol.AddBond(center_i, other_center_i, bt)
                                        new_bond = mol.GetBondBetweenAtoms(center_i, other_center_i)
                                        BondProvenance.set_bond(new_bond, 'original')
                                    break
                            else:
                                raise ValueError(f'Cannot find what {neigh} became')
        # Remove atoms
        for i in sorted(set(morituri), reverse=True):
            mol.RemoveAtom(self._get_new_index(mol, i))
        # mol.UpdatePropertyCache()
        return mol.GetMol()

    def expand_ring(self, mol: Chem.Mol) -> Chem.Mol:
        """
        Undoes collapse ring

        :param mol: untouched.
        :return:
        """
        self.journal.debug('Starting ring expansion')
        mol = Chem.RWMol(mol)
        rings = self._get_expansion_data(mol)  # List[Dict[str, List[Any]]]
        self._place_ring_atoms(mol, rings)
        # bonded_as_original. Rectifier will fix.
        self._restore_original_bonding(mol, rings)
        self.keep_copy(mol, 'Rings expanded and original bonding restored')
        self._add_novel_bonding(mol, rings)  # formerly `_ring_overlap_scenario` and `_infer_bonding_by_proximity`.
        self.keep_copy(mol, 'Rings expanded and original+novel bonding restored')
        mol.UpdatePropertyCache(strict=False)
        self._delete_collapsed(mol)
        self.keep_copy(mol, 'Rings expanded and ring-core deleted')
        mol.UpdatePropertyCache(strict=False)
        self._detriangulate(mol)
        # mol.UpdatePropertyCache()
        try:
            mol.UpdatePropertyCache(strict=False)
            mol = self._emergency_joining(mol)  # does not modify in place!
        except FragmensteinError as error:
            if self.throw_on_discard:
                raise error
            else:
                self.journal.info('Disconnect ignored due to keep_all=False')
                mol = self.get_largest_fragment(mol)
        if mol is None:
            raise ValueError('(Impossible) Failed at some point...')
        elif isinstance(mol, Chem.RWMol):
            return mol.GetMol()
        else:
            return mol

    # =========== Offset ===============================================================================================

    def offset(self, mol: Chem.Mol):
        """
        This is to prevent clashes.
        The numbers of the ori indices stored in collapsed rings are offset by the class variable (_collapsed_ring_offset)
        multiples of 100. (autoincrements to avoid dramas)

        :param mol:
        :return:
        """
        self._collapsed_ring_offset += 100
        old2new = {}
        # sort the not ringcore atoms
        for atom in mol.GetAtoms():
            if atom.GetIntProp('_ori_i') != -1:
                o = atom.GetIntProp('_ori_i')
                n = o + self._collapsed_ring_offset
                atom.SetIntProp('_ori_i', n)
                old2new[o] = n
            else:
                pass  # ringcores have -1 ori_i
        # sort the ringcore
        for atom in self._get_collapsed_atoms(mol):
            old = json.loads(atom.GetProp('_ori_is'))
            new = [i + self._collapsed_ring_offset for i in old]
            atom.SetProp('_ori_is', json.dumps(new))
            old2new = {**old2new, **dict(zip(old, new))}
        # this has to be done afterwards in case of a bonded mol
        for atom in self._get_collapsed_atoms(mol):
            old_neighss = json.loads(atom.GetProp('_neighbors'))  # if i in old2new else i
            new_neighss = [[old2new[i] for i in old_neighs if i in old2new] for old_neighs in old_neighss]
            atom.SetProp('_neighbors', json.dumps(new_neighss))
        # determine if the new atoms have close neighbours.
        pass

    def _renumber_original_indices(self, mol: Chem.Mol,
                                   mapping: Dict[int, int],
                                   name_restriction: Optional[str] = None):
        """
        Renumbers the indices in the ring virtual atom to new mapping.
        this is unused because the code needing it unfinished.

        :param mol:
        :param mapping: old index to new index dictionary
        :param name_restriction:
        :return:
        """
        for atom in mol.GetAtoms():
            if name_restriction is not None and \
                    atom.HasProp('_ori_name') and \
                    atom.GetProp('_ori_name') != name_restriction:
                continue
            i = atom.GetIntProp('_ori_i')
            if i == -1:
                # original indices
                ori = json.loads(atom.GetProp('_ori_is'))
                alt = [dd if dd not in mapping else mapping[dd] for dd in ori]
                atom.SetProp('_ori_is', json.dumps(alt))
                # original neighbors
                ori = json.loads(atom.GetProp('_neighbors'))
                alt = [[dd if dd not in mapping else mapping[dd] for dd in inner] for inner in ori]
                atom.SetProp('_neighbors', json.dumps(alt))
            elif i in mapping:
                atom.SetIntProp('_ori_i', mapping[i])
            else:
                pass

    # =========== Expand data ==========================================================================================

    def _get_expansion_data(self, mol: Chem.Mol) -> List[Dict[str, List[Any]]]:
        """
        Returns a list for each collapsed ring marking atom each with a dictionary.
        Example:

             {'atom': <rdkit.Chem.rdchem.Atom at 0x7f926fafb030>,
              'ori_name': 'APR',
              'elements': ['O', 'C', 'C', 'C', 'C'],
              'neighbors': [[4, 10], [2, 5, 6], [4, 7, 8], [6, 9, 10], [5, 8, 35]],
              'ori_is': [5, 4, 6, 8, 10],
              'xs': [0.55, 1.16, 0.785, -0.527, -0.198],
              'ys': [15.473, 15.587, 14.293, 13.909, 14.28],
              'zs': [-3.205, -4.516, -5.259, -4.577, -3.132],
              'bonds': [['SINGLE', 'SINGLE'],
               ['SINGLE', 'SINGLE', 'SINGLE'],
               ['SINGLE', 'SINGLE', 'SINGLE'],
               ['SINGLE', 'SINGLE', 'SINGLE'],
               ['SINGLE', 'SINGLE', 'SINGLE']]}

        :param mol:
        :return:
        """
        return [
            dict(atom=atom,
                 ori_name=atom.GetProp('_ori_name'),
                 elements=json.loads(atom.GetProp('_elements')),
                 neighbors=json.loads(atom.GetProp('_neighbors')),
                 ori_is=json.loads(atom.GetProp('_ori_is')),
                 xs=json.loads(atom.GetProp('_xs')),
                 ys=json.loads(atom.GetProp('_ys')),
                 zs=json.loads(atom.GetProp('_zs')),
                 bonds=json.loads(atom.GetProp('_bonds')))
            for atom in self._get_collapsed_atoms(mol)]

    def _get_expansion_for_atom(self, ring: Dict[str, List[Any]], i: int) -> Dict[str, Any]:
        """
        ``_get_expansion_data`` returns from a mol the "expansion data for the rings"
        ``_get_expansion_for_atom`` given one of the list of the data from the latter (representing a ring core)
        and an index of which of the internal atoms that were collapsed return a dictionary of details
        of that atom.
        This is called iteratively in ``_restore_original_bonding`` and ``_place_ring_atoms``.

        :param ring: see ``_get_expansion_data``
        :param i: the internal index. Say 'elements': ['C', 'C', 'C', 'O', 'C', 'C'].  i = 3 would will be Oxygen.
        :return:

        If there's something fishy use this snippet to debug:

        .. code-block:: python
            atom : Chem.Atom
            for atom in victor.monster.modifications['merged template'].GetAtoms():
                print(atom.GetIdx(), atom.GetSymbol(), atom.GetAtomicNum(), atom.GetIsotope(), atom.GetPropsAsDict(), '\n')

        """
        try:
            # ori_is and current_is
            # {'atom', 'ori_name', 'element', 'neighbor', 'ori_i', 'x', 'y', 'z', 'bond'
            return {k.replace('s',''): ring[k][i] if isinstance(ring[k], list) else ring[k] for k in ring}
        except IndexError as error:
            troublesome = [k for k in ring if isinstance(ring[k], list) and len(ring[k]) <= i]
            if len(troublesome) == 0:
                raise IndexError(f'(Ring expansion) There is a major issue with ring data for index {i} ({error}): {ring}')
            elif troublesome[0] == 'current_is':
                self.journal.warning(f'(Ring expansion) One atom {i} lacks a current index! ' + \
                                     f'This is a fallback that should not happen  ({error})')
                mol:Chem.Mol = ring['atom'].GetOwningMol()  # noqa
                ring['current_is'] = [self._get_new_index(mol, old_i, search_collapsed=False) for old_i in
                                      ring['ori_is']]
                return self._get_expansion_for_atom(ring, i)
            else:
                raise IndexError(f'The indices of the collapsed atom do not extend to {i} for {troublesome}')

    # === Key steps ====================================================================================================

    def _place_ring_atoms(self, mol, rings: List[Dict[str, Union[Chem.Atom, List[Any]]]]):
        """
        Plase all atoms stored in rings.

        :param mol:
        :param rings:
        :return:
        """
        conf = mol.GetConformer()
        for ring in rings:  # atoms in ring addition
            ringcore = ring['atom']
            indices = []  # will store current indices
            for i in range(len(ring['elements'])):  # atom addition
                collapsed_atom_data:Dict[str, Any] = self._get_expansion_for_atom(ring, i)
                # atom will be added if it is not already present!
                if self._is_present(mol, collapsed_atom_data['ori_i']):
                    natom = self._get_new_index(mol, collapsed_atom_data['ori_i'], search_collapsed=False)
                    self.journal.debug(f"{natom} (formerly {collapsed_atom_data['ori_i']} existed already: " +
                                       "fused ring or similar.")
                else:
                    n = mol.AddAtom(Chem.Atom(collapsed_atom_data['element']))
                    natom = mol.GetAtomWithIdx(n)
                    conf.SetAtomPosition(n, Point3D(collapsed_atom_data['x'],
                                                    collapsed_atom_data['y'],
                                                    collapsed_atom_data['z']))
                    natom.SetIntProp('_ori_i', collapsed_atom_data['ori_i'])
                    natom.SetDoubleProp('_x', collapsed_atom_data['x'])
                    natom.SetDoubleProp('_y', collapsed_atom_data['y'])
                    natom.SetDoubleProp('_z', collapsed_atom_data['z'])
                    natom.SetProp('_ori_name', collapsed_atom_data['ori_name'])
                    natom.SetIntProp('_ring_i', ringcore.GetIdx())
                    indices.append(n)
            ringcore.SetIntProp('_ring_i', ringcore.GetIdx())  # it really really should have not changed.
            ringcore.SetProp('_current_is', json.dumps(indices))
            ring['current_is'] = indices

    def _restore_original_bonding(self, mol: Chem.RWMol, rings: List[Dict[str, List[Any]]]) -> None:
        """
        Restore the bonding stored in the ringcore.

        The data of each collapsed atom is given by ``self._get_expansion_for_atom(ring, i)``
        """
        self.journal.debug('Restoring original bonding if any.')
        to_be_waited_for = []
        for ring in rings:
            self.journal.debug(f'Restoring ring {ring}')
            for ring_idx in range(len(ring['ori_is'])):
                # iteration per atom:
                collapsed_atom_data: Dict[str, Any] = self._get_expansion_for_atom(ring, ring_idx)
                old_i: int = collapsed_atom_data['ori_i']
                new_i: int = self._get_new_index(mol, old_i, search_collapsed=False)
                for old_neigh, bond_name in zip(collapsed_atom_data['neighbor'], collapsed_atom_data['bond']):
                    new_neigh: int = self._get_new_index(mol, old_neigh, search_collapsed=False)
                    info: str = f'ring bond #{ring_idx} between {new_i} {new_neigh}(<={old_i},{old_neigh}) '
                    self._restore_bond(mol, new_i, new_neigh, bond_name, info)

    def _restore_bond(self, mol: Chem.RWMol,
                      new_i, new_neigh, bond_name,info:str):
        """
        Called by ``_restore_original_bonding``.
        """
        # bond is string of Enum BondType.name
        bt:Chem.BondType = getattr(Chem.BondType, bond_name)
        present_bond = mol.GetBondBetweenAtoms(new_i, new_neigh)
        # The bond has not yet been restored (generally the case)
        if present_bond is None:
            assert new_i != new_neigh, f'Cannot bond to self. {info}'
            mol.AddBond(new_i, new_neigh, bt)
            present_bond: Chem.Bond = mol.GetBondBetweenAtoms(new_i, new_neigh)
            present_bond.SetBoolProp('_IsRingBond', True)
            BondProvenance.set_bond(present_bond, 'original')
            distance:float = Chem.rdMolTransforms.GetBondLength(mol.GetConformer(), new_i, new_neigh) # noqa
            assert distance < 4, f'Bond length too long ({distance}. {info}'
            self.journal.debug(f'{info} added as {bt}')
            return
        # The bond has been restored by the bond type is wrong
        present_bond_name: str = present_bond.GetBondType().name  # noqa
        if present_bond_name == bond_name:
            self.journal.debug(f'{info} exists already ' +
                               f'(has {present_bond.GetBondType().name} as expected {bt})')  # noqa
            return
        # GetBondTypeAsDouble works on bonds... not bond types
        # but there's a logic at play here:
        # aromatic trumps double, which trumps single
        preference = ['SINGLE', 'DOUBLE', 'AROMATIC']
        if preference.index(present_bond_name) > preference.index(bond_name):
            self.journal.debug(f'{info} exists already ' +
                               f'but has {present_bond.GetBondType().name} as is better than' +  # noqa
                               f'the expected {bt})')
            return
        self.journal.info(f'{info} exists already ' +
                          f'but has {present_bond.GetBondType().name} while expected {bt})')   # noqa
        present_bond.SetBondType(bt)
        present_bond.SetBoolProp('_IsRingBond', True)
        BondProvenance.set_bond(present_bond, 'original')

    def _delete_collapsed(self, mol: Chem.RWMol):
        for a in reversed(range(mol.GetNumAtoms())):
            if mol.GetAtomWithIdx(a).GetIntProp('_ori_i') == -1:
                mol.RemoveAtom(a)

    def _add_novel_bonding(self, mol: Chem.RWMol, rings: List[Dict[str, List[Any]]]):
        """
        Formerly `_ring_overlap_scenario`,`_connenct_ring`,  `_infer_bonding_by_proximity`.

        Two close rings can be one of:

        * bonded rings: Two separate rings bonded
        * Spiro rings: One atom in common
        * fused rings: Two atoms in common
        * bridged rings: counts as two rings, not three.

        The atom placement stops the same atom being placed twice...
        ...But the major issue is that the rings may and often do come from two different hits.
        Hence why it does not store it in memory.

        :param mol:
        :param rings: output of `_get_expansion_data`.
        :type rings: List[Dict[str, List[Any]]]
        :return:
        """
        self.journal.debug('Adding novel bonding (if any)...')
        # ===== Deal with Ring on ring bonding ------------------------------------------
        novel_ringcore_pairs = self._get_novel_ringcore_pairs(mol, rings, cutoff=1.5)
        # these is a list of Chem.Atom pairs.
        for ringcore_A, ringcore_B in novel_ringcore_pairs:
            self.journal.debug('determining novel bond between ring markers ' + \
                               f'{ringcore_A.GetIdx()} and {ringcore_B.GetIdx()}')
            # _determine_mergers_novel_ringcore_pair finds mergers
            self._determine_mergers_novel_ringcore_pair(mol, ringcore_A, ringcore_B)
        # ===== Deal with Ring on other bonding ------------------------------------------
        # formerly: _infer_bonding_by_proximity
        novel_other_pairs = self._get_novel_other_pairs(mol, rings, 1.0)
        for ringcore, other in novel_other_pairs:
            self.journal.debug(f'determining novel bond between ' + \
                               f'ring marker {ringcore.GetIdx()} and non-ring {other.GetIdx()}')
            # _determine_mergers_novel_ringcore_pair finds, bonds and marks for deletion.
            self._determine_mergers_novel_other_pair(mol, ringcore, other)
        # ===== Clean up ------------------------------------------------------------------
        self._delete_marked(mol)

    # =========== dependant methods =================================================================================

    def _get_new_index(self, mol: Chem.Mol, old: int, search_collapsed=True,
                       name_restriction: Optional[str] = None) -> int:
        """
        Given an old index check in ``_ori_i`` for what the current one is.
        NB. ring placeholder will be -1 and these also have ``_ori_is``. a JSON of the ori_i they summarise.

        :param mol:
        :param old: old index
        :param search_collapsed: seach also in ``_ori_is``
        :parm name_restriction: restrict to original name.
        :return:
        """
        for i, atom in enumerate(mol.GetAtoms()):
            if name_restriction is not None and atom.GetProp('_ori_name') != name_restriction:
                pass
            elif atom.GetIntProp('_ori_i') == old:
                return i
            elif search_collapsed and \
                    atom.HasProp('_ori_is') and \
                    old in json.loads(atom.GetProp('_ori_is')):
                return i
            else:
                pass
        else:
            raise ValueError(f'Cannot find {old}')

    def _is_present(self, mol, i):
        """
        Find if in ``mol`` there is an atom whose original index was ``i``.
        Wrapper around ``_get_new_index``, but does not return the index.

        :param mol:
        :param i: original index
        :return:
        """
        try:
            self._get_new_index(mol, i, search_collapsed=False)
            # raises value error.
            return True
        except ValueError as err:  # no atom is present (actually the default)
            return False

    def _get_collapsed_atoms(self, mol: Chem.Mol) -> List[Chem.Atom]:
        return [atom for atom in mol.GetAtoms() if atom.GetIntProp('_ori_i') == -1]

    # ==== Novel Ring to Ring bonding ==================================================================================

    def _get_novel_ringcore_pairs(self,
                                  mol: Chem.Mol,
                                  rings: List[Dict[str, List[Any]]],
                                  cutoff: float) \
            -> List[Tuple[Chem.Atom, Chem.Atom]]:
        """
        Get the ring atoms that are bonded or closer than a cutoff.
        It is the countepart to _get_novel_ringcore_pairs

        :param mol:
        :param rings: output of `_get_expansion_data`. See that for details.
        :type rings: List[Dict[str, List[Any]]]
        :param cutoff:
        :return:
        """
        # ----------------------------------------------------------
        # scenario where they are closer than the cutoff
        close_pairs = self._get_close_novel_ringcores(mol, rings, cutoff)  # 2.7889 ring diameter + 1.45 C-C bond
        # ----------------------------------------------------------
        # scenario where they are bonded...
        bonded_pairs = self._get_novel_ringcore_bonded_pairs(rings)
        # ------------ merge lists -----------
        return self.merge_pairing_lists(close_pairs + bonded_pairs)

    def _get_novel_ringcore_bonded_pairs(self, rings):
        """
        called by _get_novel_ringcore_pairs alongside _get_close_novel_ringcores
        Opposite of _get_novel_other_bonded_pairs

        :param rings:
        :return:
        """
        bonded_pairs = []
        for ring in rings:
            # ring: Dict[str, List[Any]]
            ringcore = ring['atom']  # Chem.Atom the ring core marker, not a real atom.
            for neigh in ringcore.GetNeighbors():
                # find those ringcore atoms that are connected. via these conditions:
                has_ringcore_neighbor = neigh.HasProp('_ori_name') and neigh.GetIntProp('_ori_i') == -1  # -1 is ring.
                is_new_pair = ringcore.GetIdx() < neigh.GetIdx()  # to avoid doing it twice each direction
                is_novel_connection = neigh.HasProp('_ori_name') and ring['ori_name'] != neigh.GetProp('_ori_name')
                # checking:
                if has_ringcore_neighbor and is_new_pair and is_novel_connection:
                    # This ringcore atom shares a novel border with another ringcore atom
                    bonded_pairs.append((ringcore, neigh))  # i.e. List[Tuple[Chem.Atom, Chem.Atom]}
        return bonded_pairs

    def merge_pairing_lists(self,
                            nonunique_pairs: List[Tuple[Chem.Atom, Chem.Atom]],
                            ringcore_first=True) \
            -> List[Tuple[Chem.Atom, Chem.Atom]]:
        # complicated because mol.GetAtomWithIdx(2) == mol.GetAtomWithIdx(2) is False
        seen = []
        pairs = []
        for atom_a, atom_b in nonunique_pairs:
            # sort out
            ai = atom_a.GetIdx()
            bi = atom_b.GetIdx()
            if ai == bi:
                self.journal.debug(f'Merging: Bond to self incident with {ai} ' +
                                   f'(ring? {atom_a.HasProp("_current_is") == 1})')
                continue
            if ringcore_first and atom_a.HasProp('_current_is') and not atom_b.HasProp('_current_is'):
                ringcore = ai
                other = bi
                pairing = f'{ringcore}-{other}'
            elif ringcore_first and atom_b.HasProp('_current_is') and not atom_a.HasProp('_current_is'):
                ringcore = bi
                other = ai
                pairing = f'{ringcore}-{other}'
                atom_a, atom_b = atom_b, atom_a
            elif atom_a.HasProp('_current_is') and atom_b.HasProp('_current_is'):
                low_i, high_i = sorted([ai, bi])
                pairing = f'{low_i}-{high_i}'
            else:
                # these should not have been let thorugh but other novel does not filter them.
                # self.journal.debug(f'Non-ring to non-ring closeness flagged!')
                continue
            # verify unseen
            if pairing in seen:
                pass
            else:
                seen.append(pairing)
                pairs.append((atom_a, atom_b))
        return pairs

    def _get_ring_atom_indices_per_origin(self, rings: List[Dict[str, List[Any]]]) -> Dict[str, List[int]]:
        atomdex = defaultdict(set)
        for ring in rings:
            origin_name = ring['ori_name']  # ring['ori_name'] is same as ringcore.GetProp('_ori_name')
            atomdex[origin_name].update(ring['current_is'])  # ring['current_is'] = json ringcore.GetProp('_current_is')
        return atomdex

    def _get_atom_indices_per_origin(self, mol: Chem.Mol) -> Dict[str, List[int]]:
        atomdex = defaultdict(list)
        for atom in mol.GetAtoms():
            if not atom.HasProp('ori_name'):
                atomdex['unknown'].append(atom.GetIdx())
            else:
                name = atom.GetProp('ori_name')
                atomdex[name].append(atom.GetIdx())
        return atomdex

    def _get_close_novel_ringcores(self, mol: Chem.Mol, rings: List[Dict[str, List[Any]]], cutoff: float):
        cnrai = self._get_close_novel_ring_atoms_indices(mol, rings, cutoff)
        return self._indices_to_atoms_n_cores(mol, cnrai)

    def _indices_to_atoms_n_cores(self, mol: Chem.Mol, index_pairs: List[Tuple[int, int]]):
        """
        Give a list of index pairs convert them to a list of pairs of atoms/cores
        """
        # these are indices of ring atoms, not ring cores
        get_atom = lambda i: mol.GetAtomWithIdx(int(i))
        get_ringcore = lambda atom: get_atom(atom.GetIntProp('_ring_i')) if atom.HasProp('_ring_i') else atom
        idx2ringcore = lambda i: get_ringcore(get_atom(i))
        return [(idx2ringcore(ai), idx2ringcore(bi)) for ai, bi in index_pairs]

    def _get_close_novel_ring_atoms_indices(self,
                                            mol: Chem.Mol,
                                            rings: List[Dict[str, List[Any]]],
                                            cutoff: float) -> List[Tuple[int, int]]:
        """
        Get the list of pairs of indices derived from a ring that are closer that ``cutoff`` to another ring atom.
        Note, the operations are between real ring atoms not ring core markers.
        Hence why ``_get_close_novel_ringcores`` does a conversion.


        :param mol:
        :param rings: output of `_get_expansion_data`. See that for details.
        :param cutoff:
        :return:
        """
        atomdex = self._get_ring_atom_indices_per_origin(rings)
        # not calling get_distance matrxi because tehre may be more than 2 origins.
        distance_matrix = Chem.Get3DDistanceMatrix(mol)
        for origin_name, indices in atomdex.items():
            # this blanks subsquares of same origin but not interesections of different indices from atomdex...
            self._nan_fill_submatrix(distance_matrix, list(indices))
        self._nan_fill_others(mol, distance_matrix, [idx for idcs in atomdex.values() for idx in idcs])
        return self._get_closest_from_matrix(distance_matrix, cutoff)

    def _get_closest_from_matrix(self, matrix: np.ndarray, cutoff: float) -> List[Tuple[int, int]]:
        # get the pair of atom indices that are less than cutoff.
        # where returns a tuple of np.arrays of dtype=np.int64
        with np.errstate(invalid='ignore'):
            return list(zip(*[w.astype(int) for w in np.where(matrix < cutoff)]))

    def _get_close_novel_ring_other_indices(self,
                                            mol: Chem.Mol,
                                            rings: List[Dict[str, List[Any]]],
                                            cutoff: int) -> List[Tuple[int, int]]:
        """
        Get the list of pairs of indices between an ring atom and a non-ring atom from a different origin that is too close.

        :param mol:
        :param rings: output of `_get_expansion_data`. See that for details.
        :param cutoff:
        :return:
        """
        atomdex = self._get_ring_atom_indices_per_origin(rings)
        oridex = self._get_atom_indices_per_origin(mol)
        distance_matrix = Chem.Get3DDistanceMatrix(mol)
        # blank all rings to rings
        self._nan_fill_submatrix(distance_matrix, [i for l in atomdex.values() for i in l])
        closeness = []
        for origin_name, indices in atomdex.items():
            sub = distance_matrix.copy()
            self._nan_fill_submatrix(sub, oridex[origin_name])
            # get the pair of atom indices that are less thna cutoff.
            # where returns a tuple of np.arrays of dtype=np.int64
            closeness.extend(self._get_closest_from_matrix(distance_matrix, cutoff))
        return closeness

    def _determine_mergers_novel_ringcore_pair(self,
                                               mol: Chem.RWMol,
                                               ringcore_A: Chem.Atom,
                                               ringcore_B: Chem.Atom) -> List[Tuple[int, int]]:
        """
        Preps to resolve ringcore pairs.
        Formerly part of ``_ring_overlap_scenario``.
        Preps without deleting. ``_delete_marked(mol)`` does that.
        bonded, spiro, fused

        :param mol:
        :param ringcore_A:
        :param ringcore_B:
        :return: list of atoms to be merged
        """
        absorption_distance = 1.  # Å
        # print('A', ringcore_A, ringcore_A.GetIdx(), ringcore_A.GetIntProp('_ori_i'))
        # print('B', ringcore_B, ringcore_B.GetIdx(), ringcore_B.GetIntProp('_ori_i'))
        indices_A = json.loads(ringcore_A.GetProp('_current_is'))
        indices_B = json.loads(ringcore_B.GetProp('_current_is'))
        distance_matrix = self._get_distance_matrix(mol, indices_A, indices_B)  # currently in `_join_neighboring`.
        # distance matrix is for the whole thing
        # TODO merge into _get_distance_matrix
        # blanking the other atom indices:
        self._nan_fill_others(mol, distance_matrix, indices_A + indices_B)
        # get closest pair.
        distance = np.nanmin(distance_matrix)
        if np.isnan(distance):
            self.journal.critical('This is impossible. Two neighbouring rings cannot be connected.')
            return []
        elif distance > absorption_distance:  # bonded
            p = np.where(distance_matrix == distance)
            a = int(p[0][0])
            b = int(p[1][0])
            present_bond = mol.GetBondBetweenAtoms(ringcore_A.GetIdx(), ringcore_B.GetIdx())
            self._add_bond_by_reference(mol, a, b, present_bond)
            self.journal.info('A novel bond-connected ring pair was found')
            self._mark_for_deletion(mol, b)
            self._copy_bonding(mol, a, b, force=True)
            return []  # bonded
        else:  # Spiro or fused.
            p = np.where(distance_matrix == distance)
            a = int(p[0][0])
            b = int(p[1][0])
            self._mark_for_deletion(mol, b)
            self._copy_bonding(mol, a, b, force=True)
            distance_matrix[a, b] = np.nan
            distance_matrix[b, a] = np.nan
            distance = np.nanmin(distance_matrix)
            if np.isnan(distance) or distance < absorption_distance:
                p = np.where(distance_matrix == distance)
                c = int(p[0][0])
                d = int(p[1][0])
                self._mark_for_deletion(mol, d)
                self._copy_bonding(mol, c, d, force=True)
                self.journal.info('A novel fused ring pair was found')
                return [(a, b), (c, d)]  # fused
            else:
                self.journal.info('A novel spiro ring pair was found')
                return [(a, b)]  # spiro

    # ==== Novel Ring to Ring bonding ==================================================================================

    # def _infer_bonding_by_proximity(self, mol):
    #     raise Exception
    #     # fix neighbours
    #     # this should not happen. But just in case!
    #     while True:
    #         ringsatoms = self._get_ring_info(mol)
    #         for ringA, ringB in itertools.combinations(ringsatoms, 2):
    #             n = mol.GetNumAtoms()
    #             self.absorb_overclose(mol, ringA, ringB, cutoff=1.)
    #             if n != mol.GetNumAtoms():
    #                 break
    #         else:
    #             break
    #     new_ringers = [atom.GetIdx() for atom in mol.GetAtoms() if atom.HasProp('expanded')]
    #     self.absorb_overclose(mol, new_ringers)
    #     new_ringers = [atom.GetIdx() for atom in mol.GetAtoms() if atom.HasProp('expanded')]
    #     self.join_overclose(mol, new_ringers)
    #     self.join_rings(mol)
    #     self._triangle_warn(mol)

    def _get_novel_other_bonded_pairs(self, rings) -> List[Tuple[Chem.Atom, Chem.Atom]]:
        pairs = []
        for ring in rings:
            # ring: Dict[str, List[Any]]
            ringcore = ring['atom']  # Chem.Atom
            # --------- find ring atom - non-origin other pairs that are bonded
            for neigh in ringcore.GetNeighbors():
                # find those ringcore-other atoms that are connected. via these conditions:
                has_notringcore_neighbor = (not neigh.HasProp('_ori_name')) or neigh.GetIntProp('_ori_i') != -1
                is_novel_connection = (not neigh.HasProp('_ori_name')) or ring['ori_name'] != neigh.GetProp('_ori_name')
                if has_notringcore_neighbor and is_novel_connection:
                    # This ringcore atom shares a novel border with another ringcore atom
                    pairs.append((ringcore, neigh))
        return pairs

    def _get_novel_other_pairs(self, mol, rings, cutoff: float) -> List[Tuple[Chem.Atom, Chem.Atom]]:
        """
        similar to self._get_novel_ringcore_pairs... but opposite
        It deals with bonded and close pairs.

        :param mol:
        :param rings: output of `_get_expansion_data`. See that for details.
        :type rings: List[Dict[str, List[Any]]]
        :param cutoff:
        :return:
        """
        stringify_pairs = lambda pairs: str([(p[0].GetIdx(), p[1].GetIdx()) for p in pairs])
        # ----------------------------------------------------------
        # scenario where they are closer than the cutoff
        close_pairs = self._get_close_novel_others(mol, rings, cutoff)  # 2.7889 ring diameter + 1.45 C-C bond
        self.journal.debug(f'_get_novel_other_pairs - close_pairs @ {cutoff}: {stringify_pairs(close_pairs)}')
        # ----------------------------------------------------------
        # scenario where they are bonded...
        bonded_pairs = self._get_novel_other_bonded_pairs(rings)
        self.journal.debug(f'_get_novel_other_pairs - bonded_pairs @ {cutoff}: {stringify_pairs(bonded_pairs)}')
        # ------------ merge lists -----------
        return self.merge_pairing_lists(close_pairs + bonded_pairs)

    def _get_close_novel_others(self,
                                mol,
                                rings,
                                cutoff):
        idx_pairs = self._get_close_novel_ring_other_indices(mol, rings, cutoff)
        # filter out identities
        idx_pairs = [(ai, bi) for (ai, bi) in idx_pairs if ai != bi]
        # filter out those that are bonded already to core atom.
        return self._indices_to_atoms_n_cores(mol, idx_pairs)

    def _get_merging_penalties(self, mol, shape: Tuple[int, int], indices_ring):
        """
        confusingly this is a different set of penalties to `_get_joining_penalties` which is for joining
        :param mol:
        :param shape:
        :param indices_ring: The indices inside the ring atom, aka. prop _current_is
        :return:
        """
        penalties = np.zeros(shape)
        for i in indices_ring:
            atom = mol.GetAtomWithIdx(i)
            neighs = [neigh for neigh in atom.GetNeighbors() if self._is_count_valid(neigh)]
            n_neighs = len(neighs)
            if atom.GetAtomicNum() > 8:  # next row.
                # weird chemistry... likely wrong!
                penalties[i, :] = 2.
                penalties[:, i] = 2.
            elif n_neighs == 2:
                pass  # no penalty!
            elif atom.GetIsAromatic():
                penalties[i, :] = 2  # this would result in a ring downgrade...
                penalties[:, i] = 2
            elif n_neighs == 3:
                penalties[i, :] = 1.  # 4 bonded carbon is not nice...
                penalties[:, i] = 1.
            else:
                penalties[i, :] = np.nan  # this will likely crash things.
                penalties[:, i] = np.nan
        return penalties

    def _get_distance(self, atom_a: Chem.Atom, atom_b: Chem.Atom) -> float:
        """
        Not sure where doing it manually is quicker than getting the whole 3D distance table.

        :param atom_a:
        :param atom_b:
        :return:
        """
        conf = atom_a.GetOwningMol().GetConformer()
        get_pos = lambda atom: np.array(conf.GetAtomPosition(atom.GetIdx()))
        return np.linalg.norm(get_pos(atom_a) - get_pos(atom_b))

    def _determine_mergers_novel_other_pair(self,
                                            mol: Chem.RWMol,
                                            ringcore: Chem.Atom,
                                            other: Chem.Atom) -> List[Tuple[int, int]]:
        """
        Like _determine_mergers_novel_ringcore_pair finds, bonds and marks for deletion.
        It however finds atoms to absorb between a ring and a given non-ring atom.

        :param mol:
        :param ringcore:
        :param other:
        :return:
        """
        # ---- Prep data.
        indices_ring = json.loads(ringcore.GetProp('_current_is'))
        index_other = other.GetIdx()
        indices_other = [index_other]
        index_core = ringcore.GetIdx()
        distance_matrix = self._get_distance_matrix(mol, indices_ring,
                                                    indices_other)  # currently in `_join_neighboring`.
        self._nan_fill_others(mol, distance_matrix, indices_ring + indices_other)
        # merging penalties
        penalties = self._get_merging_penalties(mol, distance_matrix.shape, indices_ring)
        core_absorption_distance = 1.5  # Å between ring core and other. 2.8 Å is diameter.
        core_other_distance = self._get_distance(ringcore, other)
        if core_other_distance < core_absorption_distance:
            # ------ Within ring must go. No penalties.
            self.journal.debug(
                f'(DetMergeNovOther *{index_core}, {index_other}). Other is within ring. Forcing absoption')
            absorption_distance = 9999
            pendist_matrix = distance_matrix
        else:
            # ------ Assess cases normally.
            absorption_distance = 1.  # Å between ring atom and other.
            pendist_matrix = penalties + distance_matrix
        # get closest pair.
        pendistance = np.nanmin(pendist_matrix)
        if np.isnan(pendistance):
            self.journal.warning(f'(DetMergeNovOther*{index_core}, {index_other}). This is impossible...')
            return []
        else:  # bonded
            p = np.where(pendist_matrix == pendistance)
            a = int(p[0][0])
            b = int(p[1][0])
            assert index_other in (a, b), 'CRITICIAL: Matrix error!'
            # absorb or bond
            distance = distance_matrix[a, b]
            penalty = penalties[a, b]  # penalties were already applied. this is for msgs only
            if distance > 4:
                self.journal.warning(f'(DetMergeNovOther: {a}, {b}). ' + \
                                     f'The bond between {a} and {b} too long {distance} ' + \
                                     f'from {indices_ring} and {indices_other}')
                return []
            elif distance > absorption_distance:
                self.journal.info(f'(DetMergeNovOther: {a}, {b}). A novel bonding to ring may be added. ' + \
                                  f'd: {distance} p: {penalty}')
                # get bond type
                present_bond = mol.GetBondBetweenAtoms(ringcore.GetIdx(), other.GetIdx())
                self._add_bond_by_reference(mol, a, b, present_bond)
            else:
                # absorb the non-ring atom!
                self.journal.info(f'(DetMergeNovOther: {a}, {b}). An atom was absorbed to ring. ' + \
                                  f'd: {distance} p: {penalty}')
                if mol.GetAtomWithIdx(a).GetIntProp('_ori_i') == -1:
                    self._copy_bonding(mol, a, b)
                    self._mark_for_deletion(mol, b)
                    return [(a, b)]
                else:
                    self._copy_bonding(mol, b, a)
                    self._mark_for_deletion(mol, a)
                    return [(b, a)]

    def _add_bond_by_reference(self, mol, a, b, reference_bond):
        """
        _copy_bonding copies all the bonds. THis just adds one like the reference bond.
        It calls ``_add_bond_if_possible`` if its a closeness bond, i.e. reference_bond is None
        It calls ``_add_bond_regardlessly`` if its an orginal one.

        :param mol:
        :param a:
        :param b:
        :param reference_bond:
        :return:
        """
        atom_a = mol.GetAtomWithIdx(a)
        atom_b = mol.GetAtomWithIdx(b)
        if reference_bond is None:
            self._add_bond_if_possible(mol, atom_a, atom_b, 'other_novel')
        else:
            bt = reference_bond.GetBondType()
            if bt is None or bt == Chem.BondType.UNSPECIFIED:
                bt = Chem.BondType.SINGLE
            self._add_bond_regardlessly(mol, atom_a, atom_b, bt, BondProvenance.get_bond(reference_bond).name)

    # ======== Emergency ===============================================================================================

    def _emergency_joining(self, mol: Chem.Mol) -> Chem.Mol:
        self.journal.debug('`_emergency_joining` called')
        return self._join_internally(mol, severe=True)

    def _join_internally(self, mol: Chem.Mol, severe: bool = False) -> Chem.Mol:
        """
        The last check to see if the mol is connected.
        This differs (and calls) ``join_neighboring_mols``

        """
        is_rw = isinstance(mol, Chem.RWMol)
        frags = Chem.GetMolFrags(mol, asMols=True, sanitizeFrags=False)
        n = len(frags)
        if n == 1:
            return mol
        else:
            while n > 1:
                if severe:
                    self.journal.info(f'Molecule disconnected in {n} parts. Please inspect final product!')
                else:
                    self.journal.debug('Linking two disconnected fragments')
                # ----- get names ---------------
                name = mol.GetProp('_Name')
                for i, frag in enumerate(frags):
                    frag.SetProp('_Name', f'name.{i}')
                # find which fragments are closest ------------------------------
                # TODO use the distance_matrix = self._get_distance_matrix(..) code
                closeness = np.ones([n, n])
                closeness.fill(float('nan'))
                for a, b in itertools.combinations(list(range(n)), 2):
                    closeness[a, b] = self._find_closest(frags[a], frags[b])[3]
                p = np.where(closeness == np.nanmin(closeness))
                frags = list(frags)
                first = frags[p[0][0]]
                second = frags[p[1][0]]
                mol = self.join_neighboring_mols(first, second)
                frags.remove(first)
                frags.remove(second)
                # ---- reset variables ---------
                for part in frags:
                    mol = Chem.CombineMols(mol, part)
                    mol.SetProp('_Name', name)
                frags = Chem.GetMolFrags(mol, asMols=True, sanitizeFrags=False)
                n = len(frags)
            if is_rw:
                return Chem.RWMol(mol)
            else:
                return mol

        #
        #
        #     distance = distance_matrix[anchor_A, anchor_B]
        #
        #
        # rpos = np.array(conf.GetAtomPosition(ringcore_A.GetIdx()))
        # npos = np.array(conf.GetAtomPosition(ringcore_B.GetIdx()))
        # if np.linalg.norm(rpos - npos) <= 4:
        #     pairs = []
        #     for G_idxs, ref_i in [(A_idxs, neigh.GetIdx()), (B_idxs, ringcore.GetIdx())]:
        #         tm = np.take(dm, G_idxs, 0)
        #         tm2 = np.take(tm, [ref_i], 1)
        #         p = np.where(tm2 == np.nanmin(tm2))
        #         f = G_idxs[int(p[0][0])]
        #         tm2[p[0][0], :] = np.ones(tm2.shape[1]) * float('inf')
        #         p = np.where(tm2 == np.nanmin(tm2))
        #         s = G_idxs[int(p[1][0])]
        #         pairs.append((f, s))
        #     # now determine which are closer
        #     if dm[pairs[0][0], pairs[1][0]] < dm[pairs[0][0], pairs[1][1]]:
        #         mergituri.append((pairs[0][0], pairs[1][0]))
        #         mergituri.append((pairs[0][1], pairs[1][1]))
        #     else:
        #         mergituri.append((pairs[0][0], pairs[1][1]))
        #         mergituri.append((pairs[0][1], pairs[1][0]))

    def _detriangulate(self, mol: Chem.RWMol) -> None:
        """
        Prevents novel cyclopropanes and cyclobutanes.

        :param mol:
        :return:
        """
        for atom in mol.GetAtoms():
            atom_i = atom.GetIdx()
            for neigh in atom.GetNeighbors():
                neigh_i = neigh.GetIdx()
                if neigh_i < atom_i:  # dont check twice...
                    continue
                # de triangulate?
                third_i = self._get_triangle(atom, neigh)
                if third_i is not None:
                    # it is a triangle
                    third = mol.GetAtomWithIdx(third_i)
                    self.journal.debug(f'Triangle present {(atom_i, neigh_i, third_i)}.')
                    combinator = partial(itertools.combinations, r=2)
                    combinator.__qualname__ = 'triangle'
                    self._detriangulate_inner(mol,
                                              atoms=[atom, neigh, third],
                                              atom_indices=[atom_i, neigh_i, third_i],
                                              combinator=combinator
                                              )
                    # start again
                    return self._detriangulate(mol)
                # de square-ify
                sq = self._get_square(atom, neigh)
                if sq is not None:
                    far_i, close_i = sq
                    far = mol.GetAtomWithIdx(far_i)
                    close = mol.GetAtomWithIdx(close_i)  # second neighbour of atom
                    # bonding is:
                    # atom - neigh - far - close - atom
                    self.journal.debug(f'Square present {(atom_i, neigh_i, far_i, close_i)}.')
                    # combinations would fail at diagonally opposite angles as they arent bonded
                    # the order is irrelevant if the same fun is called
                    combinator = lambda l: [[l[0], l[1]],
                                            [l[0], l[2]],
                                            [l[1], l[3]],
                                            [l[2], l[3]]]
                    combinator.__qualname__ = 'quadrangle'
                    self._detriangulate_inner(mol,
                                              atoms=[atom, neigh, close, far],
                                              atom_indices=[atom_i, neigh_i, close_i, far_i],
                                              combinator=combinator
                                              )

    def _detriangulate_inner(self,
                             mol: Chem.RWMol,
                             atoms: List[Chem.Atom],
                             atom_indices: List[int],
                             combinator: Callable):
        """
        Triangle/square agnostic.

        :param mol:
        :param atoms:
        :param atom_indices:
        :param combinator:
        :return:
        """
        bonds = [mol.GetBondBetweenAtoms(a, b) for a, b in combinator(atom_indices)]
        if any([bond is None for bond in bonds]):
            # this bond was fixed in the previous iteration
            # self.journal.critical(f'IMPOSSIBLE ERROR: detriangulate missing bond ' +
            #                       f'(combinator={combinator.__qualname__}. ' +
            #                       f'{atom_indices})')
            return None
        provenances = BondProvenance.get_bonds(bonds)
        # original
        originality = [p == BondProvenance.ORIGINAL for p in provenances]
        if sum(originality) == 3:
            self.journal.debug('Triangle from original present. Kept, unless rectifiers is set to not tolerate')
        # length
        BL = partial(Chem.rdMolTransforms.GetBondLength, conf=mol.GetConformer())
        lengths = [BL(iAtomId=b.GetBeginAtomIdx(), jAtomId=b.GetEndAtomIdx()) for b in bonds]
        scores = np.array(lengths)
        scores[originality] = np.nan
        # atom properties. overbonded, ring etc.
        for fun, weight in self.closeness_weights:
            funscore = [fun(a) for a in atoms]
            scores += np.sum(list(combinator(funscore)), axis=1)
        d = np.nanmax(scores)
        if np.isnan(d):
            return None  # impossible but okay.
        doomed_i = int(np.where(scores == d)[0])
        doomed_bond = bonds[doomed_i]
        a, b = doomed_bond.GetBeginAtomIdx(), doomed_bond.GetEndAtomIdx()
        self.journal.debug(f'Removing triangle/square forming bond between {a} and {b}')
        mol.RemoveBond(a, b)



# Testing discarding
from rdkit import Chem
from fragmenstein import Walton, Victor, Igor
import unittest
from fragmenstein import DistanceError

class TestDiscard(unittest.TestCase):
    pdb_block = \
        '''ATOM      1  N   ALA A   1      -0.677  -1.230  -0.491  1.00  0.00           N  
        ATOM      2  CA  ALA A   1      -0.001   0.064  -0.491  1.00  0.00           C  
        ATOM      3  C   ALA A   1       1.499  -0.110  -0.491  1.00  0.00           C  
        ATOM      4  O   ALA A   1       2.030  -1.227  -0.502  1.00  0.00           O  
        ATOM      5  CB  ALA A   1      -0.509   0.856   0.727  1.00  0.00           C  
        ATOM      6  H   ALA A   1      -0.131  -2.162  -0.491  1.00  0.00           H  
        ATOM      7  HA  ALA A   1      -0.269   0.603  -1.418  1.00  0.00           H  
        ATOM      8 1HB  ALA A   1      -1.605   1.006   0.691  1.00  0.00           H  
        ATOM      9 2HB  ALA A   1      -0.285   0.342   1.681  1.00  0.00           H  
        ATOM     10 3HB  ALA A   1      -0.053   1.861   0.784  1.00  0.00           H  
        TER   
        END'''

    def setUp(self) -> None:
        Igor.init_pyrosetta()

    def get_separated(self) -> Victor:
        wally = Walton.from_smiles(benzene='c1ccccc1', tetrazole='[nH]1nncc1')
        wally.ring_on_plane(ring_idx=0, mol_idx=0)
        wally.superpose_by_mcs()
        wally.translate_parallel(mol_idx=1,
                                 distance=10,
                                 base_atom_idx=0,
                                 pointer_atom_idx=2)
        Victor.monster_throw_on_discard = True
        vicky = Victor(wally.mols,
                       pdb_block=self.pdb_block,
                       ligand_resn='LIG')
        return vicky
    def test_discard(self):
        vicky: Victor = self.get_separated()
        try:
            vicky.combine(joining_cutoff=3)
            print(vicky.summarize())
            self.fail('Should have thrown')
        except DistanceError:
            print('error thrown as expected')

    def test_success(self):
        vicky: Victor = self.get_separated()
        vicky.monster.throw_on_discard = True
        vicky.combine(joining_cutoff=12)
        print(vicky.summarize())

    def test_success2(self):
        vicky: Victor = self.get_separated()
        vicky.place('c1nnn(CCCCCOc2ccccc2)c1')
        print(vicky.summarize())

# these test need a bit more meat:
# the second benzene is not a contributor as all overlap.
# wally = Walton.from_smiles(benzene='c1ccccc1', benzene2='c1ccccc1')
# wally.ring_on_plane(ring_idx=0, mol_idx=0)
# wally.superpose_by_mcs()
# vicky = Victor(wally.mols,
#                pdb_block=pdb_block,
#                ligand_resn='LIG')
# # vicky.monster_throw_on_discard = True <-- NO!
# vicky.monster.throw_on_discard = True
# vicky.place('c1cccc(CCCCCOc2ccccc2)c1', long_name='discard-error2')




import functools

from rdkit.Chem import rdFMCS
from ..error import FragmensteinError
from ..victor import Victor
from rdkit import Chem, Geometry
from rdkit.Chem import AllChem


class Quicktor(Victor):
    """
    This class is intended to speed up the Victor placement step, by:

    * doing a strict MCS only
    * not allowing constrained atoms to move during RDKit minimisation and killing them if they fail.
    * doing a quick pyrosetta minimisation

    Example usage:

    .. code-block:: python
       lab = Laboratory(hits, pdb_block=pdb_block)
       lab.Victor = Quicktor
       to_be_placed: pd.DataFrame = place_input_validator(to_be_placed)
       placed: pd.DataFrame = lab.place(to_be_placed_df, n_cores=55, expand_isomers=False, timeout=60)
    """
    quick_reanimation = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.monster.matching_modes = [{'atomCompare': rdFMCS.AtomCompare.CompareElements,
                                        'bondCompare': rdFMCS.BondCompare.CompareOrder,
                                        'ringCompare': rdFMCS.RingCompare.PermissiveRingFusion,
                                        'ringMatchesRingOnly': True}]
        self.monster_mmff_minisation = True  # as is default. Repeated for emphasis.

    @functools.cached_property
    def preminimized_undummied_mol(self) -> Chem.Mol:
        """
        This method is called by the plonking into structure methods.
        Not "positioning" as intended by ``monster`` is done.
        Opening a PDB in RDKit is doable but gets exponentially slow with chain length
        """
        mol = Chem.Mol(self.monster.positioned_mol)
        # ## Store xyz as properties
        atom: Chem.Atom
        conf: Chem.Conformer = mol.GetConformer()  # noqa
        for atom in mol.GetAtoms():  # noqa
            if atom.HasProp('_x'):
                continue
            xyz: Geometry.Point3D = conf.GetAtomPosition( atom.GetIdx() )  # noqa
            atom.SetDoubleProp('_x', float(xyz.x))
            atom.SetDoubleProp('_y', float(xyz.y))
            atom.SetDoubleProp('_z', float(xyz.z))
        # ## Minimise
        if self.monster_mmff_minisation:
            self.journal.debug(f'{self.long_name} - pre-minimising monster (MMFF)')
            min_result = self.monster.mmff_minimize(mol, allow_lax=False, ff_max_displacement=float('nan'))
            if not min_result.success:
                raise FragmensteinError('Could not preminize')
            ddG = min_result.delta
            if ddG > 100:
                raise FragmensteinError('Poor preminization')
        return AllChem.DeleteSubstructs(min_result.mol, Chem.MolFromSmiles('*'))
